__author__ = "Yuri E. Corilo"
__date__ = "Jul 31, 2019"

from copy import deepcopy
from threading import Thread
from numpy import average, std
from corems.molecular_id.calc.ClusterFilter import ClusteringFilter
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulas
from corems.molecular_id.factory.molecularSQL import MolForm_SQL 
from tqdm import tqdm

class FindOxygenPeaks(Thread):
    
    '''
        Class to walk 14Da units over oxygen space for negative ion mass spectrum of natural organic matter
        Returns a list of MSPeak class containing the possible Molecular Formula class objects.  
        
        Parameters
        ----------
        mass_spectrum_obj : MassSpec class
            This is where we store MassSpec class obj,   
        
        lookupTableSettings:  MolecularLookupTableSettings class
            This is where we store MolecularLookupTableSettings class obj
        
        min_O , max_O : int
            minium and maximum of Oxygen to allow the software to look for
            it will override the settings at lookupTableSettings.usedAtoms
            default min = 1, max = 30

        Attributes
        ----------
        mass_spectrum_obj : MassSpec class
            This is where we store MassSpec class obj,   
        lookupTableSettings:  MolecularLookupTableSettings class
            This is where we store MolecularLookupTableSettings class obj
        
        Methods
        ----------
            run()    
                will be called when the instantiated class method start is called
            get_list_found_peaks()
                returns a list of MSpeaks classes cotaining all the MolecularFormula candidates inside the MSPeak
                for more details of the structure see MSPeak class and MolecularFormula class    
            set_mass_spec_indexes_by_found_peaks()
                set the mass spectrum to interate over only the selected indexes
    '''
    def __init__(self, mass_spectrum_obj, sql_db=False, min_O = 1, max_O = 22) :
        
        Thread.__init__(self)
        
        self.mass_spectrum_obj = mass_spectrum_obj
        self.min_0 = min_O
        self.max_O = max_O
        
        if not sql_db:
            
            self.sql_db = MolForm_SQL(mass_spectrum_obj.molecular_search_settings.url_database)
        else:

            self.sql_db = sql_db    
    
    def run(self):
        
        #save initial settings min peaks per class filter 
        initial_min_peak_bool = deepcopy(self.mass_spectrum_obj.molecular_search_settings.use_min_peaks_filter)

        #deactivate the usage of min peaks per class filter
        self.mass_spectrum_obj.molecular_search_settings.use_min_peaks_filter = False

        #save initial settings for Ox 
        initial_ox = deepcopy(self.mass_spectrum_obj.molecular_search_settings.usedAtoms['O'])

        #resets the used atoms to look only for oxygen organic compounds
        self.mass_spectrum_obj.molecular_search_settings.usedAtoms['O'] =  (self.min_0, self.max_O)
        
        self.list_found_mspeaks = []

        kdm_base = self.mass_spectrum_obj.mspeaks_settings.kendrick_base
        
        self.mass_spectrum_obj.change_kendrick_base_all_mspeaks(kdm_base)
        
        # needs to be wrapped inside the mass_spec class
        ClusteringFilter().filter_kendrick(self.mass_spectrum_obj)
        
        print("Start most abundant mass spectral peak search") 
        molecular_formula_obj_reference = self.find_most_abundant_formula(self.mass_spectrum_obj)
        
        print("Select most abundant peak with molecular formula =  %s with a m/z error of %s ppm" % (molecular_formula_obj_reference.string, molecular_formula_obj_reference.mz_error))
        
        print("Started mass spectral peak series search")

        self.list_found_mspeaks = self.find_series_mspeaks(self.mass_spectrum_obj,
                                                           molecular_formula_obj_reference, 
                                                           deltamz=14)
        
        # reset indexes after done with operation that includes a filter (i.e. ClusteringFilter().filter_kendrick())
        
        self.mass_spectrum_obj.molecular_search_settings.usedAtoms['O'] =  initial_ox
        
        self.mass_spectrum_obj.molecular_search_settings.use_min_peaks_filter = initial_min_peak_bool
        
        self.mass_spectrum_obj.reset_indexes()

        self.mass_spectrum_obj.filter_by_noise_threshold()
        
        print("Done with mass spectral peak series search")

        self.sql_db.close()

    def find_most_abundant_formula(self, mass_spectrum_obj):
        '''
        find most abundant using kendrick 
        
        Returns
        ----------
        MolecularFormula class obj
            most abundant MolecularFormula with the lowest mass error
        '''
        #need to find a better way to cut off outliners
        #import matplotlib.pyplot as plt
        #plt.hist(mass_spectrum_obj.abundance, bins=100)
        #plt.show()
        
        abundances =  mass_spectrum_obj.abundance
        abun_mean = average(abundances, axis=0)
        abun_std = std(abundances, axis=0)
        
        upper_limit = abun_mean + 7* abun_std
        
        print("Maximum abundance limit  = %s and max abundance kendrick cluster = %s"  % (upper_limit, max(mass_spectrum_obj, key=lambda m: m.abundance).abundance))
        
        mspeak_most_abundant = max(mass_spectrum_obj, key=lambda m: m.abundance if m.abundance <= upper_limit else 0)

        print("Searching molecular formulas")

        SearchMolecularFormulas(mass_spectrum_obj, self.sql_db).run_worker_ms_peaks([mspeak_most_abundant])
        
        print("Finished searching molecular formulas")

        if mspeak_most_abundant:

            return mspeak_most_abundant.best_molecular_formula_candidate
        
        else:
        
            raise Exception("Could not find a possible molecular formula match for the most abundant peak of m/z %.5f"%mspeak_most_abundant.mz_exp )
        
        #return the first option
        #return mspeak_most_abundant[0]

    def find_most_abundant_formula_test(self, mass_spectrum_obj, settings):
        
        #this function is intended for test only. 
        # Have to sort by Kendrick to be able to select the most abundant series 
        #then select the most abundant peak inside the series
        #or have the user select the reference mspeak on the gui

        mspeak_most_abundant = mass_spectrum_obj.most_abundant_mspeak

        SearchMolecularFormulas(mass_spectrum_obj, self.sql_db).run_worker_ms_peaks([mspeak_most_abundant])
        
        if mspeak_most_abundant:

            return mspeak_most_abundant.best_molecular_formula_candidate 
            
        else:
            raise Exception("Could not find a possible molecular formula match for the most abundant peak of m/z %.5f"%mspeak_most_abundant.mz_exp )
        #return the first option
        #return mspeak_most_abundant[0]
    
    def find_series_mspeaks(self, mass_spectrum_obj, molecular_formula_obj_reference, deltamz=14):

        abundances =  mass_spectrum_obj.abundance
        abun_mean = average(abundances, axis=0)
        abun_std = std(abundances, axis=0)
        upper_limit = abun_mean + 7* abun_std
       
        list_most_abundant_peaks = list()

        min_mz = mass_spectrum_obj.min_mz_exp
        
        max_mz = mass_spectrum_obj.max_mz_exp
        
        initial_nominal_mass = molecular_formula_obj_reference.mz_nominal_calc
        
        mass = initial_nominal_mass
        
        nominal_masses = []
        while mass <= max_mz:
            #print "shit 1", mass, min_mz
            mass += (deltamz) 
            nominal_masses.append(mass)
        
        mass = initial_nominal_mass    
        while mass >= min_mz:
            #print "shit 1", mass, min_mz
            mass -= (deltamz) 
            nominal_masses.append(mass)
        
        nominal_masses = sorted(nominal_masses)
        
        for nominal_mass in nominal_masses:
            
            first_index, last_index = mass_spectrum_obj.get_nominal_mz_first_last_indexes(nominal_mass)
            
            ms_peaks = mass_spectrum_obj[first_index:last_index]
            
            if ms_peaks:   
                '''    
                print (nominal_mass, first_index, 
                    last_index, 
                    mass_spectrum_obj[first_index].mz_exp,
                    mass_spectrum_obj[last_index].mz_exp
                    )
                '''
                
                mspeak_most_abundant = max(ms_peaks, key=lambda m: m.abundance if m.abundance <= upper_limit else 0)

                #mspeak_most_abundant = max(ms_peaks, key=lambda m: m.abundance)
                
                list_most_abundant_peaks.append(mspeak_most_abundant)
        
        print('Start molecular formula search')
        SearchMolecularFormulas(mass_spectrum_obj, self.sql_db).run_worker_ms_peaks(list_most_abundant_peaks)
        print('Done molecular formula search')
        return [mspeak for mspeak in list_most_abundant_peaks if mspeak]            
                
    
    def get_list_found_peaks(self):
        
        return sorted(self.list_found_mspeaks, key=lambda mp: mp.mz_exp)

    def set_mass_spec_indexes_by_found_peaks(self):
        
        '''
        Warning!!!!
        set the mass spectrum to interate over only the selected indexes
        don not forget to call mass_spectrum_obj.reset_indexes after the job is done
        '''
        
        indexes = [msp.index for msp in self.list_found_mspeaks]
        self.mass_spectrum_obj.set_indexes(indexes)
        
        



from sklearn.cluster import DBSCAN
from sklearn.preprocessing import StandardScaler
from sklearn.cluster import MeanShift, estimate_bandwidth



import numpy as np
import pandas as pd
#import matplotlib.pyplot as plt

class ClusteringFilter():
    
    def get_mass_error_matrix_data(self, ms_peaks):
        
        mass_list = list()
        error_list = list()
        list_indexes_mass_spec = []
        
        for index, mspeak in enumerate(ms_peaks):

            if mspeak.is_assigned:
                    
                #print(mspeak.mz_exp, len(mspeak))
                for mformula in mspeak:
                    mass_list.append(mspeak.mz_exp)
                    error_list.append(mformula.mz_error)
                    list_indexes_mass_spec.append(index)
        
        kendrick_dict = {'mass': mass_list, 'error': error_list}  
        df = pd.DataFrame(kendrick_dict) 
        matrix_data = df.values.astype("float32", copy = False)
        return matrix_data, list_indexes_mass_spec

    def get_kendrick_matrix_data(self, mass_spectrum):
        
        km = mass_spectrum.kendrick_mass
        kdm = mass_spectrum.kmd
        kendrick_dict = {'km': km, 'kdm': kdm}  
        df = pd.DataFrame(kendrick_dict) 
        matrix_data = df.values.astype("float32", copy = False)
        return matrix_data

    def filter_kendrick(self, mass_spectrum):
        
        matrix_data = self.get_kendrick_matrix_data(mass_spectrum)

        stdscaler = StandardScaler().fit(matrix_data)
        
        matrix_data_scaled = stdscaler.transform(matrix_data)

        clusters = DBSCAN(eps = .75, min_samples=50).fit_predict(matrix_data_scaled)
        
        # Number of clusters in labels, ignoring noise if present.
        n_clusters_ = len(set(clusters)) - (1 if -1 in clusters else 0)
        n_noise_ = list(clusters).count(-1)
        
        indexes = []
        for i in range(len(clusters)):
            if clusters[i] == -1:
                indexes.append(i)
        
        print('Estimated number of clusters: %d' % n_clusters_)
        print('Estimated number of noise points: %d' % n_noise_)
        print()
        mass_spectrum.filter_by_index(indexes)
        #from matplotlib import pyplot as plt
        #plt.scatter(matrix_data[:, 0], matrix_data[:, 1], c=clusters, cmap="jet")
        #plt.xlabel("km")
        #plt.ylabel("kdm")
        #plt.show()
        #plt.close()

    def filter_kendrick_by_index(self, ms_peak_indexes, mass_spectrum_obj):
        
        min_samples = mass_spectrum_obj.molecular_search_settings.min_peaks_per_class

        kendrick_dict = {'km': list(), 'kmd': list()}  

        if len(ms_peak_indexes) <= 1: return []
        
        for index, _ in ms_peak_indexes:
           kendrick_dict["km"].append(mass_spectrum_obj[index].kendrick_mass)
           kendrick_dict["kmd"].append(mass_spectrum_obj[index].kmd)
           
        # check min data points otherwise StandardScaler().fit(0 will fail
        
        df = pd.DataFrame(kendrick_dict) 
        matrix_data = df.values.astype("float32", copy = False)

        stdscaler = StandardScaler().fit(matrix_data)
        matrix_data_scaled = stdscaler.transform(matrix_data)

        clusters = DBSCAN(eps = .8, min_samples=min_samples).fit_predict(matrix_data_scaled)
        
        # Number of clusters in labels, ignoring noise if present.
        n_clusters_ = len(set(clusters)) - (1 if -1 in clusters else 0)
        n_noise_ = list(clusters).count(-1)
        
        print('Estimated number of clusters: %d' % n_clusters_)
        print('Estimated number of noise points: %d' % n_noise_)
        print()

        noise_idx = []
        
        other_peaks_idx = []

        for i in range(len(clusters)):
            
            if clusters[i] == -1:
                noise_idx.append(ms_peak_indexes[i])
            
            else:
                other_peaks_idx.append(ms_peak_indexes[i])    

        #mfs = [mass_spectrum_obj[index].best_molecular_formula_candidate.string for index in other_peaks_idx]
        
        #mfs_noise = [mass_spectrum_obj[index].best_molecular_formula_candidate.string for index in noise_idx]
        
        #print(mfs)
        #print(mfs_noise)

        #from matplotlib import pyplot as plt
        #plt.scatter(matrix_data[:, 0], matrix_data[:, 1], c=clusters, cmap="jet")
        #plt.xlabel("km")
        #plt.ylabel("kdm")
        #plt.show()
        #plt.close()
        
        return noise_idx      

    def remove_assignment_by_mass_error(self, mass_spectrum):
        
        #data need to be binned by mz unit or more to be able to use clustering
        
        matrix_data, list_indexes_mass_spec = self.get_mass_error_matrix_data(mass_spectrum)

        stdscaler = StandardScaler().fit(matrix_data)
        
        matrix_data_scaled = stdscaler.transform(matrix_data)
        
        #bandwidth = estimate_bandwidth(matrix_data_scaled, quantile=0.3, n_samples=int(len(ms_peaks)/3))

        #clusters = MeanShift(bandwidth=bandwidth, bin_seeding=True).fit_predict(matrix_data_scaled)
        
        #eps and min_samp need to be optimized by precision and number of mspeaks
        clusters = DBSCAN(eps = .15).fit_predict(matrix_data_scaled)
        
        indexes = []
        
        #from matplotlib import pyplot as plt
        #plt.scatter(matrix_data[:, 0], matrix_data[:, 1], c=clusters, cmap="plasma")
        #plt.xlabel("km")
        #plt.ylabel("kdm")
        #plt.show()
        #plt.close()

        for i in range(len(clusters)):
            if clusters[i] == -1:
                indexes.append(list_indexes_mass_spec[i])
        
        mass_spectrum.remove_assignment_by_index(indexes)    


__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"

import sys
sys.path.append('.')
from pathlib import Path
from threading import Thread
from corems.mass_spectrum.input.massList import ReadCoremsMasslist
from corems.mass_spectra.factory.LC_Class import LCMSBase

class ReadCoremsMassSpectraText(ReadCoremsMasslist, Thread):
    
    def __init__(self, file_location, analyzer='Unknown', instrument_label='Unknown'):
        
        """
         # Parameters
		----------
        file_location: text,  pathlib.Path(), or s3path.S3Path 
            Path object from pathlib containing the file location
        """
        
        if  isinstance(file_location, str):
			# if obj is a string it defaults to create a Path obj, pass the S3Path if needed
            file_location = Path(file_location)

        if not file_location.exists():
            raise FileNotFoundError("%s not found" % file_location)
        
        if not file_location.suffix == '.corems':
            
            raise TypeError("%s is not a valid CoreMS file" % file_location)
        
        Thread.__init__(self)
        
        ReadCoremsMasslist.__init__(self, file_location)
        
        self.lcms = LCMSBase(self.file_location, analyzer=analyzer,instrument_label=instrument_label)

    def get_scans_filename(self):
        
        all_other = self.file_location.glob('*_scan*[!.json]')

        scans_filepath = [(file_path_obj.stem.split('scan')[1], file_path_obj)  for file_path_obj in all_other]
        
        scans_filepath.sort(key=lambda m: int(m[0]))

        return scans_filepath
        
    def set_filepath_datatype_and_delimiter(self, file_path_obj):

        self.file_location = file_path_obj

        if file_path_obj.suffix == '.pkl':

            self.data_type == 'dataframe'
            
        else:

            if file_path_obj.suffix == '.csv':
                self.data_type == 'txt'    
                self.delimiter = ','    
            
            elif file_path_obj.suffix == '.xlsx':
                self.data_type == 'excel'    
                self.delimiter = ',' 
            
            elif file_path_obj.suffix == '.txt':
                self.data_type == 'txt'    
                self.delimiter = '\t'
                print('WARNING using tab as delimiter')     
            else:
                raise NotImplementedError('%s data not yet supported ' % file_path_obj.suffix)
     
    def import_mass_spectra(self):
        
        list_rt, list_tic, list_scan = list(), list(), list()
        
        for scan_number, file_path_obj in self.get_scans_filename():
            
            self.set_filepath_datatype_and_delimiter(file_path_obj)
            
            mass_spec = self.get_mass_spectrum(int(scan_number))

            list_scan.append(int(scan_number))

            list_rt.append(mass_spec.rt)

            list_tic.append(mass_spec.tic)
            
            self.lcms.add_mass_spectrum(mass_spec)

        self.lcms.retention_time = list_rt
        self.lcms.tic_list = list_tic
        self.lcms.scans_number = list_scan
     
    def run(self):
        '''creates the lcms obj'''

        self.import_mass_spectra()
            
    def get_lcms_obj(self):
        
        if self.lcms:
            
            return self.lcms
        
        else:
            
            raise Exception("returning a empty lcms class")

if __name__ == "__main__":
    
    file_location = Path.cwd() / "NEG_ESI_SRFA_CoreMS.corems"
    
    all_other = file_location.glob('*_scan*[!.json]')

    all_json = file_location.glob('*_scan*.json')

    names_scans = [filename.stem.split('scan')  for filename in all_other]
    
    print(names_scans[0]) 
    #names_scans.sort(key=lambda m: int(m[1]))
    
    


'''
@author: Yuri E. Corilo
@date: Jun 27, 2019
'''

from numpy import hstack, inf, isnan, poly1d, polyfit, where
from corems.encapsulation.constant import Labels

class PeakPicking:

    def cut_mz_domain_peak_picking(self):

        max_picking_mz = self.settings.max_picking_mz
        min_picking_mz = self.settings.min_picking_mz

        min_final = where(self.mz_exp_profile > min_picking_mz)[-1][-1]
        min_comeco = where(self.mz_exp_profile > min_picking_mz)[0][0]

        mz_domain_X_low_cutoff, mz_domain_low_Y_cutoff, = self.mz_exp_profile[min_comeco: min_final], self.abundance_profile[min_comeco: min_final]

        max_final = where(self.mz_exp_profile < max_picking_mz)[-1][-1]
        max_comeco = where(self.mz_exp_profile < max_picking_mz)[0][0]

        if self.has_frequency:

            if self.freq_exp_profile.any():

                freq_domain_low_Y_cutoff = self.freq_exp_profile[min_comeco:min_final]

                return mz_domain_X_low_cutoff[max_comeco:max_final], mz_domain_low_Y_cutoff[max_comeco:max_final], freq_domain_low_Y_cutoff[max_comeco:max_final]

        else:

            return mz_domain_X_low_cutoff[max_comeco:max_final], mz_domain_low_Y_cutoff[max_comeco:max_final], None

    def do_peak_picking(self):

        mz, abudance, freq = self.cut_mz_domain_peak_picking()

        if self.label == Labels.bruker_frequency or self.label == Labels.midas_frequency:

            self.calc_centroid(mz, abudance, freq)

        elif self.label == Labels.thermo_profile:
            self.calc_centroid(mz, abudance, self.freq_exp_profile)

        elif self.label == Labels.bruker_profile:
            self.calc_centroid(mz, abudance, self.freq_exp_profile)

        elif self.label == Labels.booster_profile:
            self.calc_centroid(mz, abudance, self.freq_exp_profile)

        elif self.label == Labels.simulated_profile:
            self.calc_centroid(mz, abudance, self.freq_exp_profile)

        else: raise Exception("Unknow mass spectrum type", self.label)

    def find_minima(self, apex_index, abundance, len_abundance, right=True):
            
            j = apex_index
            
            if right: minima = abundance[j] > abundance[j+1]
            else: minima = abundance[j] > abundance[j-1]

            while minima:
                
                if j == 1 or j == len_abundance -2:
                    break
                
                if right: 
                    j += 1

                    minima = abundance[j] >= abundance[j+1]

                else: 
                    j -= 1
                    minima = abundance[j] >= abundance[j-1]
            
            if right: return j
            else: return j

    def calculate_resolving_power(self, intes, massa, current_index):
            
            '''this is a conservative calculation of resolving power,
               the peak need to be resolved at least at the half-maximum magnitude,
               otherwise, the combined full width at half maximum is used to calculate resolving power'''

            peak_height = intes[current_index]
            target_peak_height = peak_height/2

            peak_height_minus = peak_height
            peak_height_plus = peak_height

            index_minus = current_index
            while peak_height_minus  >= target_peak_height:

                index_minus = index_minus -1
                peak_height_minus = intes[index_minus]
                #print "massa", "," , "intes", "," , massa[index_minus], "," , peak_height_minus
            x = [ massa[index_minus],  massa[index_minus+1]]
            y = [ intes[index_minus],  intes[index_minus+1]]
            coefficients = polyfit(x, y, 1)

            a = coefficients[0]
            b = coefficients[1]

            y_intercept =  intes[index_minus] + ((intes[index_minus+1] - intes[index_minus])/2)
            massa1 = (y_intercept -b)/a

            index_plus = current_index
            while peak_height_plus  >= target_peak_height:

                index_plus = index_plus + 1
                peak_height_plus = intes[index_plus]
                #print "massa", "," , "intes", "," , massa[index_plus], "," , peak_height_plus

            x = [massa[index_plus],  massa[index_plus - 1]]
            y = [intes[index_plus],  intes[index_plus - 1]]

            coefficients = polyfit(x, y, 1)
            a = coefficients[0]
            b = coefficients[1]

            y_intercept =  intes[index_plus - 1] + ((intes[index_plus] - intes[index_plus - 1])/2)
            massa2 = (y_intercept -b)/a

            if massa1 > massa2:

                resolvingpower =  massa[current_index]/(massa1-massa2)

            else:

                resolvingpower =  massa[current_index]/(massa2-massa1)

            return resolvingpower

    def cal_minima(self, mass, abun):

        abun = -abun

        dy = abun[1:] - abun[:-1]
        
        '''replaces nan for infinity'''
        indices_nan = where(isnan(abun))[0]
        
        if indices_nan.size:
            
            abun[indices_nan] = inf
            dy[where(isnan(dy))[0]] = inf
        
        indexes = where((hstack((dy, 0)) < 0) & (hstack((0, dy)) > 0))[0]

        if indexes.size:
            
            return mass[indexes], abun[indexes]
    
    def calc_centroid(self, mass, abund, freq):
        #TODO: remove peaks that minimum is one data point from the maximum
        # to remove artifacts 

        len_abundance = len(abund)
        
        max_abundance = max(abund)
        
        peak_height_diff = lambda hi, li : ((abund[hi] - abund[li]) / max_abundance )*100

        abundance_threshold, factor = self.get_threshold(abund)
        #print(abundance_threshold, factor)
        # find indices of all peaks
        dy = abund[1:] - abund[:-1]
        
        '''replaces nan for infinity'''
        indices_nan = where(isnan(abund))[0]
        
        if indices_nan.size:
            
            abund[indices_nan] = inf
            dy[where(isnan(dy))[0]] = inf
        
        indexes = where((hstack((dy, 0)) < 0) & (hstack((0, dy)) > 0))[0]
        
        # noise threshold
        if indexes.size and abundance_threshold is not None:
            indexes = indexes[abund[indexes]/factor >= abundance_threshold]
            
        for current_index in indexes: 
            
            if self.label == Labels.simulated_profile: 

                mz_exp_centroid, intes_centr, peak_indexes = self.use_the_max(mass, abund, current_index, len_abundance, peak_height_diff)
                if mz_exp_centroid:
                    
                    peak_resolving_power = self.calculate_resolving_power( abund, mass, current_index)
                    s2n = intes_centr/self.baselise_noise_std
                    freq_centr = None
                    self.add_mspeak(self.polarity, mz_exp_centroid, abund[current_index] , peak_resolving_power, s2n, peak_indexes, exp_freq=freq_centr, ms_parent=self)
            
            else:
            
                mz_exp_centroid, freq_centr, intes_centr, peak_indexes = self.find_apex_fit_quadratic(mass, abund, freq, current_index, len_abundance, peak_height_diff)
                if mz_exp_centroid:
                    
                    peak_resolving_power = self.calculate_resolving_power( abund, mass, current_index)
                    s2n = intes_centr/self.baselise_noise_std
                    self.add_mspeak(self.polarity, mz_exp_centroid, abund[current_index] , peak_resolving_power, s2n, peak_indexes, exp_freq=freq_centr, ms_parent=self)
            
        
    def get_threshold(self, intes):
        
        threshold_method = self.settings.threshold_method

        if threshold_method == 'auto':
            
            #print(self.settings.noise_threshold_std)
            abundance_threshold = self.baselise_noise + (self.settings.noise_threshold_std * self.baselise_noise_std)
            factor = 1

        elif threshold_method == 'signal_noise':

            abundance_threshold = self.settings.s2n_threshold
            factor = self.baselise_noise_std

        elif threshold_method == "relative_abundance":

            abundance_threshold = self.settings.relative_abundance_threshold
            factor = intes.max()/100

        else:
            raise  Exception("%s method was not implemented, please refer to corems.mass_spectrum.calc.NoiseCalc Class" % threshold_method)
        
        return abundance_threshold, factor
        
    def find_apex_fit_quadratic(self, mass, abund, freq, current_index, len_abundance, peak_height_diff):
        
        # calc prominence
        peak_indexes = self.check_prominence(abund, current_index, len_abundance, peak_height_diff )
        
        if not peak_indexes:        
            
            return None, None, None, None           
        
        else:    
            
            # fit parabola to three most abundant datapoints
            list_mass = [mass[current_index - 1], mass[current_index], mass[current_index +1]]
            list_y = [abund[current_index - 1],abund[current_index], abund[current_index +1]]
            
            z = poly1d(polyfit(list_mass, list_y, 2))
            a = z[2]
            b = z[1]

            calculated = -b/(2*a)
            
            if calculated < 1 or int(calculated) != int(list_mass[1]):

                mz_exp_centroid = list_mass[1]
            
            else:
                
                mz_exp_centroid = calculated 
            
            if self.label == Labels.bruker_frequency or self.label == Labels.midas_frequency:
                
                # fit parabola to three most abundant frequency datapoints
                list_freq = [freq[current_index - 1], freq[current_index], freq[current_index +1]]
                z = poly1d(polyfit(list_freq, list_y, 2))
                a = z[2]
                b = z[1]

                calculated_freq = -b/(2*a)

                if calculated_freq < 1 or int(calculated_freq) != freq[current_index]:
                    freq_centr = list_freq[1]

                else:
                    freq_centr = calculated_freq
            
            else:
                    freq_centr = None
                    
            return mz_exp_centroid, freq_centr, abund[current_index], peak_indexes
    
    def check_prominence(self, abun, current_index, len_abundance, peak_height_diff ):

        final_index = self.find_minima(current_index, abun, len_abundance, right=True)
            
        start_index = self.find_minima(current_index, abun, len_abundance, right=False)
            
        peak_indexes = (start_index, current_index, final_index)

        if min( peak_height_diff(current_index,start_index), peak_height_diff(current_index,final_index) ) >  self.mspeaks_settings.peak_min_prominence_percent :   
            
            return peak_indexes
        
        else:
            
            return False

    def use_the_max(self, mass, abund, current_index, len_abundance, peak_height_diff):

        peak_indexes = self.check_prominence(abund, current_index, len_abundance, peak_height_diff )
        
        if not peak_indexes:        

            return None, None, None
        
        else:    
            
            return mass[current_index], abund[current_index], peak_indexes


                    


from pathlib import Path
from copy import deepcopy


#from matplotlib import rcParamsDefault, rcParams
from numpy import array, power, float64, where

from corems.mass_spectrum.calc.MassSpectrumCalc import MassSpecCalc
from corems.mass_spectrum.calc.KendrickGroup import KendrickGrouping
from corems.encapsulation.constant import Labels
from corems.ms_peak.factory.MSPeakClasses import ICRMassPeak as MSPeak
from corems.encapsulation.factory.parameters import MSParameters
from corems.encapsulation.input.parameter_from_json import load_and_set_parameters_ms

__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"

def overrides(interface_class):
    def overrider(method):
        assert method.__name__ in dir(interface_class)
        return method
    return overrider

class MassSpecBase(MassSpecCalc, KendrickGrouping):
    '''
    - A iterative mass spectrum base class, stores the profile data and instrument settings
    - Iteration over a list of MSPeaks classes stored at the _mspeaks attributes
    - _mspeaks is populated under the hood by calling process_mass_spec method
    - iteration is null if _mspeaks is empty

    Parameters
    ----------
    mz_exp : list(float)
        list containing the imported experiemental m/z 
        (default is to store profile mode data, but it depends on the input type)
    abundance: list(float)
        list containing the imported abundance 
        (default is to store profile mode data, but it depends on the input type)
    d_params : dict{'str': float,int and str}
        The keyword arguments are used for ...

    Attributes
    ----------
    _mz_exp : list(float)
        This is where we store mz_exp,
    _abundance : list(float)     
        This is where we store _abundance,
    _mspeaks : list(MSPeak)
        store MSpeaks objects identified by a peak picking algorithm

    Relevant Methods
    ----------
    process_mass_spec()
        find or set the noise threshold base on the setting encapsulated at settings.input.ProcessingSetting.MassSpectrumSetting
        - run the peak peaking algorithm and use the method addMSPeaks() to populate _mspeaks attribute

    see also: MassSpecCentroid(), MassSpecfromFreq(), MassSpecProfile()
    '''
    def __init__(self, mz_exp, abundance, d_params, **kwargs):

        self._abundance = array(abundance, dtype=float64)
        self._mz_exp = array(mz_exp, dtype=float64)

        # objects created after process_mass_spec() function
        self._mspeaks = list()
        self._dict_nominal_masses_indexes = dict()
        self._baselise_noise = 0.001
        self._baselise_noise_std = 0.001
        self._dynamic_range = None
        # set to None: initialization occurs inside subclass MassSpecfromFreq
        self._transient_settings = None
        self._frequency_domain = None
        self._mz_cal_profile = None
        self.is_calibrated = False

        self._set_parameters_objects(d_params)
        self._init_settings()

        self.is_centroid = False
        self.has_frequency = False

        self.calibration_order = None
        self.calibration_points = None
        self.calibration_RMS = None

    def _init_settings(self):

        self._parameters = MSParameters()

    def __len__(self):

        return len(self.mspeaks)

    def __getitem__(self, position):

        return self.mspeaks[position]

    def set_indexes(self, list_indexes):
        ''' set the mass spectrum to interate over only the selected MSpeaks indexes'''
        self.mspeaks = [self._mspeaks[i] for i in list_indexes]

        for i, mspeak in  enumerate(self.mspeaks): mspeak.index = i

        self._set_nominal_masses_start_final_indexes()

    def reset_indexes(self):
        ''' reset the mass spectrum to interate over all MSpeaks obj'''
        self.mspeaks = self._mspeaks

        for i, mspeak in  enumerate(self.mspeaks): mspeak.index = i

        self._set_nominal_masses_start_final_indexes()

    def add_mspeak(self, ion_charge, mz_exp,
                    abundance,
                    resolving_power,
                    signal_to_noise,
                    massspec_indexes,
                    exp_freq=None,
                    ms_parent=None
                ):

        mspeak = MSPeak(
                ion_charge,
                mz_exp,
                abundance,
                resolving_power,
                signal_to_noise,
                massspec_indexes,
                len(self._mspeaks),
                exp_freq=exp_freq,
                ms_parent=ms_parent,
        )

        self._mspeaks.append(mspeak)

    def _set_parameters_objects(self, d_params):

        self._calibration_terms = (
            d_params.get("Aterm"),
            d_params.get("Bterm"),
            d_params.get("Cterm"),
        )

        self.label = d_params.get(Labels.label)

        self.analyzer = d_params.get('analyzer')

        self.instrument_label = d_params.get('instrument_label')

        self.polarity = int(d_params.get("polarity"))

        self.scan_number = d_params.get("scan_number")

        self.rt = d_params.get("rt")

        self.mobility_rt = d_params.get("mobility_rt")

        self.mobility_scan = d_params.get("mobility_scan")

        self._filename = d_params.get("filename_path")

        self._dir_location = d_params.get("dir_location")

        self._baselise_noise = d_params.get("baselise_noise")

        self._baselise_noise_std = d_params.get("baselise_noise_std")

        if d_params.get('sample_name') != 'Unknown':

            self.sample_name = d_params.get('sample_name')
            if not self.sample_name:
                self.sample_name = self.filename.stem
        else:

            self.sample_name = self.filename.stem

    def reset_cal_therms(self, Aterm, Bterm, C, fas= 0):

        self._calibration_terms = (Aterm, Bterm, C)

        self._mz_exp = self._f_to_mz()
        self._abundance = self._abundance
        self.find_peaks()
        self.reset_indexes()
        # self.reset_indexes()

    def clear_molecular_formulas(self):

        self.check_mspeaks()
        return array([mspeak.clear_molecular_formulas() for mspeak in self.mspeaks])

    def process_mass_spec(self, keep_profile=True, auto_noise=True, noise_bayes_est=False):
        
        # if runned mannually make sure to rerun filter_by_noise_threshold     
        # calculates noise threshold 
        # do peak picking( create mspeak_objs) 
        # reset mspeak_obj the indexes
         
        self.cal_noise_threshold(auto=auto_noise, bayes=noise_bayes_est)

        self.find_peaks()
        self.reset_indexes()

        if self.mspeaks:
            self._dynamic_range = self.max_abundance / self.min_abundance
        else:
            self._dynamic_range = 0
        if not keep_profile:

            self._abundance *= 0
            self._mz_exp *= 0
            

    def cal_noise_threshold(self, auto=True, bayes=False):

        if self.label == Labels.simulated_profile:

            self._baselise_noise, self._baselise_noise_std = 0.1, 1

        else:

            self._baselise_noise, self._baselise_noise_std = self.run_noise_threshold_calc(auto, bayes=bayes)
        
    @property
    def parameters(self):
        return self._parameters

    @parameters.setter
    def parameters(self, instance_MSParameters):
        self._parameters = instance_MSParameters

    def set_parameter_from_json(self, parameters_path):
        load_and_set_parameters_ms(self, parameters_path=parameters_path)    

    @property
    def mspeaks_settings(self): return self.parameters.ms_peak

    @mspeaks_settings.setter
    def mspeaks_settings(self, instance_MassSpecPeakSetting):

            self.parameters.ms_peak = instance_MassSpecPeakSetting

    @property
    def settings(self): return self.parameters.mass_spectrum

    @settings.setter
    def settings(self, instance_MassSpectrumSetting):

        self.parameters.mass_spectrum =  instance_MassSpectrumSetting

    @property
    def molecular_search_settings(self):  return self.parameters.molecular_search

    @molecular_search_settings.setter
    def molecular_search_settings(self, instance_MolecularFormulaSearchSettings):

        self.parameters.molecular_search =  instance_MolecularFormulaSearchSettings

    @property
    def freq_exp_profile(self):
        return self._frequency_domain

    @property
    def mz_cal_profile(self):

        return self._mz_cal_profile

    @mz_cal_profile.setter
    def mz_cal_profile(self, mz_cal_list):

        if len(mz_cal_list) == len(self.mz_exp_profile):
            self._mz_cal_profile = mz_cal_list
        else:
            raise Exception( "calibrated array (%i) is not of the same size of the data (%i)" % (len(mz_cal_list),  len(self.mz_exp_profile)))    

    @property
    def mz_cal(self):
        return array([mspeak.mz_cal for mspeak in self.mspeaks])

    @mz_cal.setter
    def mz_cal(self, mz_cal_list):

            if  len(mz_cal_list) == len(self.mspeaks):
                self.is_calibrated = True
                for index, mz_cal in enumerate(mz_cal_list):
                    self.mspeaks[index].mz_cal = mz_cal
            else: 
                raise Exception( "calibrated array (%i) is not of the same size of the data (%i)" % (len(mz_cal_list),  len(self._mspeaks)))    

    @property
    def mz_exp(self):

        self.check_mspeaks()

        if self.is_calibrated:

            return array([mspeak.mz_cal for mspeak in self.mspeaks])

        else:

            return array([mspeak.mz_exp for mspeak in self.mspeaks])

    @property
    def mz_exp_profile(self): return self._mz_exp

    @mz_exp_profile.setter
    def mz_exp_profile(self, _mz_exp ): self._mz_exp = _mz_exp

    @property
    def abundance_profile(self): return self._abundance

    @abundance_profile.setter
    def abundance_profile(self, _abundance): return self._abundance

    @property
    def abundance(self):
        self.check_mspeaks()
        return array([mspeak.abundance for mspeak in self.mspeaks])

    def freq_exp(self):
        self.check_mspeaks()
        return array([mspeak.freq_exp for mspeak in self.mspeaks])

    @property
    def resolving_power(self):
        self.check_mspeaks()
        return array([mspeak.resolving_power for mspeak in self.mspeaks])

    @property
    def signal_to_noise(self):
        self.check_mspeaks()
        return array([mspeak.signal_to_noise for mspeak in self.mspeaks])

    @property
    def nominal_mz(self):

        if self._dict_nominal_masses_indexes:

            return sorted(list(self._dict_nominal_masses_indexes.keys()))

        else:

            raise ValueError("Nominal indexes not yet set")    

    def get_mz_and_abundance_peaks_tuples(self):

        self.check_mspeaks()
        return [(mspeak.mz_exp, mspeak.abundance) for mspeak in self.mspeaks]

    @property
    def kmd(self):
        self.check_mspeaks()
        return array([mspeak.kmd for mspeak in self.mspeaks])

    @property
    def kendrick_mass(self):
        self.check_mspeaks()
        return array([mspeak.kendrick_mass for mspeak in self.mspeaks])

    @property
    def max_mz_exp(self):
        return max([mspeak.mz_exp for mspeak in self.mspeaks])

    @property
    def min_mz_exp(self):
        return min([mspeak.mz_exp for mspeak in self.mspeaks])

    @property
    def max_abundance(self):
        
        return max([mspeak.abundance for mspeak in self.mspeaks])

    @property
    def max_signal_to_noise(self):
        return max([mspeak.signal_to_noise for mspeak in self.mspeaks])

    @property
    def most_abundant_mspeak(self):
        return max(self.mspeaks, key=lambda m: m.abundance)

    @property
    def min_abundance(self):
        return min([mspeak.abundance for mspeak in self.mspeaks])

    # takes too much cpu time 
    @property
    def dynamic_range(self):
        return self._dynamic_range

    @property
    def baselise_noise(self):
        if self._baselise_noise:
            return self._baselise_noise
        else:     
            return None

    @property
    def baselise_noise_std(self):
        if self._baselise_noise_std:
            return self._baselise_noise_std
        else:     
            return None

    @property
    def Aterm(self):
        return self._calibration_terms[0]

    @property
    def Bterm(self):
        return self._calibration_terms[1]

    @property
    def Cterm(self):
        return self._calibration_terms[2]


    @property
    def filename(self):
        return Path(self._filename)

    @property
    def dir_location(self):
        return self._dir_location

    def sort_by_mz(self):
        return sorted(self, key=lambda m: m.mz_exp)

    def sort_by_abundance(self, reverse=False):
        return sorted(self, key=lambda m: m.abundance, reverse=reverse)

    @property
    def tic(self):

        return sum(self.abundance_profile)

    def check_mspeaks_warning(self):
        import warnings
        if self.mspeaks:
            pass
        else:
            warnings.warn(
                "mspeaks list is empty, continuing without filtering data"
            )

    def check_mspeaks(self):
        if self.mspeaks:
            pass
        else:
            raise Exception(
                "mspeaks list is empty, please run process_mass_spec() first"
            )

    def remove_assignment_by_index(self, indexes):
        for i in indexes: self.mspeaks[i].clear_molecular_formulas()

    def filter_by_index(self, list_indexes):

        self.mspeaks = [self.mspeaks[i] for i in range(len(self.mspeaks)) if i not in list_indexes]

        for i, mspeak in  enumerate(self.mspeaks): mspeak.index = i

        self._set_nominal_masses_start_final_indexes()

    def filter_by_mz(self, min_mz, max_mz):

        self.check_mspeaks_warning()
        indexes = [index for index, mspeak in enumerate(self.mspeaks) if min_mz <= mspeak.mz_exp <= max_mz]
        self.filter_by_index(indexes)

    def filter_by_s2n(self, min_s2n, max_s2n=False):

        self.check_mspeaks_warning()
        if not max_s2n:
            max_s2n = self.max_signal_to_noise

        self.check_mspeaks_warning()
        indexes = [index for index, mspeak in enumerate(self.mspeaks) if min_s2n <= mspeak.signal_to_noise <= max_s2n ]
        self.filter_by_index(indexes)

    def filter_by_abundance(self, min_abund, max_abund=False):

        self.check_mspeaks_warning()
        if not max_abund:
            max_abund = self.max_abundance
        indexes = [index for index, mspeak in enumerate(self.mspeaks) if min_abund <= mspeak.abundance <= max_abund]
        self.filter_by_index(indexes)

    def filter_by_max_resolving_power(self, B, T):

        rpe = lambda m, z: (1.274e7 * z * B * T)/(m*z)

        self.check_mspeaks_warning()

        indexes_to_remove = [index for index, mspeak in enumerate(self.mspeaks) if  mspeak.resolving_power >= rpe(mspeak.mz_exp,mspeak.ion_charge)]
        self.filter_by_index(indexes_to_remove)

    def filter_by_min_resolving_power(self, B, T):

        rpe = lambda m, z: (1.274e7 * z * B * T)/(m*z)

        self.check_mspeaks_warning()

        indexes_to_remove = [index for index, mspeak in enumerate(self.mspeaks) if  mspeak.resolving_power <= rpe(mspeak.mz_exp,mspeak.ion_charge)]
        self.filter_by_index(indexes_to_remove)

    def filter_by_noise_threshold(self):
        
        threshold = self.get_noise_threshold()[1][0]
        
        self.check_mspeaks_warning()
        
        indexes_to_remove = [index for index, mspeak in enumerate(self.mspeaks) if  mspeak.abundance <= threshold]
        self.filter_by_index(indexes_to_remove)

    
    def find_peaks(self):
        """needs to clear previous results from peak_picking"""
        self._mspeaks = list()
        """then do peak picking"""

        self.do_peak_picking()
        # print("A total of %i peaks were found" % len(self._mspeaks))

    def change_kendrick_base_all_mspeaks(self, kendrick_dict_base):
        """kendrick_dict_base = {"C": 1, "H": 2} or {{"C": 1, "H": 1, "O":1} etc """

        self.parameters.ms_peak.kendrick_base = kendrick_dict_base

        for mspeak in self.mspeaks:

            mspeak.change_kendrick_base(kendrick_dict_base)

    def get_nominal_mz_first_last_indexes(self, nominal_mass):

        if self._dict_nominal_masses_indexes:

            if nominal_mass in self._dict_nominal_masses_indexes.keys():

                return (self._dict_nominal_masses_indexes.get(nominal_mass)[0], self._dict_nominal_masses_indexes.get(nominal_mass)[1]+1)

            else:
                # import warnings
                # uncomment warn to distribution
                # warnings.warn("Nominal mass not found in _dict_nominal_masses_indexes, returning (0, 0) for nominal mass %i"%nominal_mass)
                return (0, 0)
        else:
            raise Exception("run process_mass_spec() function before trying to access the data")

    def get_masses_count_by_nominal_mass(self):

        dict_nominal_masses_count = {}

        all_nominal_masses = list(set([i.nominal_mz_exp for i in self.mspeaks]))

        for nominal_mass in all_nominal_masses:
            if nominal_mass not in dict_nominal_masses_count:
                dict_nominal_masses_count[nominal_mass] = len(list(self.get_nominal_mass_indexes(nominal_mass)))

        return dict_nominal_masses_count

    def datapoints_count_by_nominal_mz(self, mz_overlay=0.1):

        dict_nominal_masses_count ={}

        all_nominal_masses = list(set([i.nominal_mz_exp for i in self.mspeaks]))

        for nominal_mass in all_nominal_masses:

            if nominal_mass not in dict_nominal_masses_count:

                min_mz = nominal_mass - mz_overlay

                max_mz = nominal_mass + 1 + mz_overlay

                indexes = indexes = where((self.mz_exp_profile > min_mz) & (self.mz_exp_profile < max_mz)) 

                dict_nominal_masses_count[nominal_mass] = indexes[0].size

        return dict_nominal_masses_count

    def get_nominal_mass_indexes(self, nominal_mass, overlay=0.1):

        min_mz_to_look = nominal_mass - overlay
        max_mz_to_look = nominal_mass + 1 + overlay

        return (i for i in range(len(self.mspeaks)) if min_mz_to_look <= self.mspeaks[i].mz_exp <= max_mz_to_look)

        # indexes = (i for i in range(len(self.mspeaks)) if min_mz_to_look <= self.mspeaks[i].mz_exp <= max_mz_to_look)
        # return indexes

    def _set_nominal_masses_start_final_indexes(self):

        '''return ms peaks objs indexes(start and end) on the mass spectrum for all nominal masses'''
        dict_nominal_masses_indexes ={}

        all_nominal_masses = set(i.nominal_mz_exp for i in self.mspeaks)

        for nominal_mass in all_nominal_masses:

            indexes = self.get_nominal_mass_indexes(nominal_mass)

            defaultvalue = None
            first = last = next(indexes, defaultvalue)
            for last in indexes:
                pass

            dict_nominal_masses_indexes[nominal_mass] = (first, last)

        self._dict_nominal_masses_indexes = dict_nominal_masses_indexes

    def plot_centroid(self, ax=None, c='g'):

        import matplotlib.pyplot as plt
        if self._mspeaks:

            if ax is None:
                ax = plt.gca()

            markerline_a, stemlines_a, baseline_a = ax.stem(self.mz_exp, self.abundance, linefmt='-', markerfmt=" ", use_line_collection=True)

            plt.setp(markerline_a, 'color', c, 'linewidth', 2)
            plt.setp(stemlines_a, 'color', c, 'linewidth', 2)
            plt.setp(baseline_a, 'color', c, 'linewidth', 2)

            ax.set_xlabel("$\t{m/z}$", fontsize=12)
            ax.set_ylabel('Abundance', fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=12)

            ax.axes.spines['top'].set_visible(False)
            ax.axes.spines['right'].set_visible(False)

            ax.get_yaxis().set_visible(False)
            ax.spines['left'].set_visible(False)

        else:

            raise Exception("No centroid data found, please run process_mass_spec")

        return ax


    def plot_profile_and_noise_threshold(self, ax=None): 

        import matplotlib.pyplot as plt
        if self.baselise_noise_std and self.baselise_noise_std:

            # x = (self.mz_exp_profile.min(), self.mz_exp_profile.max())
            baseline = (self.baselise_noise, self.baselise_noise)

            # std = self.parameters.mass_spectrum.noise_threshold_std
            # threshold = self.baselise_noise_std + (std * self.baselise_noise_std)
            x, y = self.get_noise_threshold()    
            
            if ax is None:
                ax = plt.gca()
            
            ax.plot(self.mz_exp_profile, self.abundance_profile, color="green")
            ax.plot(x, (baseline, baseline), color="yellow")
            ax.plot(x, y, color="red")

            ax.set_xlabel("$\t{m/z}$", fontsize=12)
            ax.set_ylabel('Abundance', fontsize=12)
            ax.tick_params(axis='both', which='major', labelsize=12)

            ax.axes.spines['top'].set_visible(False)
            ax.axes.spines['right'].set_visible(False)

            ax.get_yaxis().set_visible(False)
            ax.spines['left'].set_visible(False)

        else:

            raise Exception("Calculate noise threshold first")

        return ax

    def plot_mz_domain_profile(self, color='green', ax=None): #pragma: no cover

        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()
        ax.plot(self.mz_exp_profile, self.abundance_profile, color=color)
        ax.set(xlabel='m/z', ylabel='abundance')

        return ax

    def to_excel(self, out_file_path):

        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(out_file_path, self)
        exportMS.to_excel()

    def to_hdf(self, out_file_path):
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(out_file_path, self)
        exportMS.to_hdf()

    def to_csv(self, out_file_path):
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(out_file_path, self)
        exportMS.to_csv()

    def to_pandas(self, out_file_path):
        # pickle dataframe (pkl extension)
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(out_file_path, self)
        exportMS.to_pandas()

    def to_dataframe(self,):
        # returns pandas dataframe

        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(self.filename, self)
        return exportMS.get_pandas_df()

    def to_json(self):
        # returns pandas dataframe

        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(self.filename, self)
        return exportMS.to_json()

    def parameters_json(self):

        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS = HighResMassSpecExport(self.filename, self)
        return exportMS.parameters_to_json()

class MassSpecProfile(MassSpecBase):
    '''
    - A iterative mass spectrum class when the entry point is on profile format
    - Stores the profile data and instrument settings
    - Iteration over a list of MSPeaks classes stored at the _mspeaks attributes
    - _mspeaks is populated under the hood by calling process_mass_spec method
    - iteration is null if _mspeaks is empty

    Parameters
    ----------
    dataframe : pandas Dataframe(Series(floats))
        contains columns [m/z, Abundance, Resolving Power, S/N] 
    d_params : dict{'str': float, int or str}

    Attributes
    ----------
    _mz_exp : list(float)
        This is where we store mz_exp,
    _abundance : list(float)     
        This is where we store _abundance,
    _mspeaks : list(MSPeak)
        store MSpeaks objects identified by a peak picking algorithm

    Relevant Methods
    ----------
    process_mass_spec()
        find or set the noise threshold base on the setting encapsulated at settings.input.ProcessingSetting.MassSpectrumSetting
        - run the peak peaking algorithm and use the method addMSPeaks() to populate _mspeaks attribute

    see also: MassSpecBase(), MassSpecfromFreq(), MassSpecProfile()
    '''

    def __init__(self, data_dict, d_params, auto_process=True, auto_noise=True, noise_bayes_est=False):
        """
        method docs
        """
        # print(data_dict.keys())
        super().__init__(data_dict.get(Labels.mz), data_dict.get(Labels.abundance), d_params)
       
        if auto_process:
            self.process_mass_spec(auto_noise=auto_noise, noise_bayes_est=noise_bayes_est)

class MassSpecfromFreq(MassSpecBase):
    '''
    - A iterative mass spectrum class when data entry is on frequency(Hz) domain
    - Transform to m/z based on the settings stored at d_params
    - Stores the profile data and instrument settings
    - Iteration over a list of MSPeaks classes stored at the _mspeaks attributes
    - _mspeaks is populated under the hood by calling process_mass_spec method
    - iteration is null if _mspeaks is empty

    Parameters
    ----------
    frequency_domain : list(float)
        all datapoints in frequency domain in Hz
    magnitude :  frequency_domain : list(float)
        all datapoints in for magnitude of each frequency datapoint

    Attributes
    ----------
    _mz_exp : list(float)
        This is where we store mz_exp,
    _frequency_domain : list(float)
        This is where we store _frequency_domain,
    _abundance : list(float)     
        This is where we store _abundance,
    _mspeaks : list(MSPeak)
        store MSpeaks objects identified by a peak picking algorithm     
    label : str
        store label (Bruker, Midas Transient, see Labels class ). It across distinct processing points
    
    Relevant Methods
    ----------
    _set_mz_domain()
        calculates the m_z based on the setting of d_params

    process_mass_spec()
        find or set the noise threshold base on the setting encapsulated at settings.input.ProcessingSetting.MassSpectrumSetting
        - run the peak peaking algorithm and use the method addMSPeaks() to populate _mspeaks attribute
    
    see also: MassSpecBase(), MassSpecfromFreq(), MassSpecProfile()
    '''

    def __init__(self, frequency_domain, magnitude, d_params, 
                auto_process=True, keep_profile=True, auto_noise=False, noise_bayes_est=False):
        """
        method docs
        """
        super().__init__(None, magnitude, d_params)

        self._frequency_domain = frequency_domain
        self.has_frequency = True
        self._set_mz_domain()

        
        """ use this call to automatically process data as the object is created, Setting need to be changed before initiating the class to be in effect"""
        
        if auto_process:
            self.process_mass_spec(keep_profile=keep_profile, auto_noise=auto_noise, noise_bayes_est=noise_bayes_est)


    def _set_mz_domain(self):

        if self.label == Labels.bruker_frequency:

            self._mz_exp = self._f_to_mz_bruker()

        else:

            self._mz_exp = self._f_to_mz()

    @property
    def transient_settings(self): return self.parameters.transient

    @transient_settings.setter
    def transient_settings(self, instance_TransientSetting):
     
        self.parameters.transient = instance_TransientSetting  


class MassSpecCentroid(MassSpecBase):

    '''
    - A iterative mass spectrum class when data entry is centroid mode
    - Stores the centroid data and instrument settings
    - Simulate profile data based on Gaussian or Lorentzian peak shape
    - Iteration over a list of MSPeaks classes stored at the _mspeaks attributes
    - _mspeaks is populated under the hood by calling process_mass_spec method
    - iteration is null if _mspeaks is empty

    Parameters
    ----------
    data_dict : dict {string: numpy array float64 )
        contains keys [m/z, Abundance, Resolving Power, S/N] 
    d_params : dict{'str': float, int or str}
        
    Attributes
    ----------
    _mz_exp : list(float)
        This is where we store mz_exp,
    _abundance : list(float)     
        This is where we store _abundance,
    _mspeaks : list(MSPeak)
        store MSpeaks objects identified by a peak picking algorithm  

    Attributes
    ----------
    _mz_exp : list(float)
        This is where we store mz_exp,
    _frequency_domain : list(float)
        This is where we store _frequency_domain,
    _abundance : list(float)     
        This is where we store _abundance,
    _mspeaks : list(MSPeak)
        store MSpeaks objects identified by a peak picking algorithm     
    label : str
        store label (Bruker, Midas Transient, see Labels class)
    
    Relevant Methods
    ----------
    _set_mz_domain()
        calculates the m_z based on the setting of d_params

    process_mass_spec()
        - overrides the base class function
        - Populates _mspeaks list with MSpeaks class using the centroid date

    see also: MassSpecBase(), MassSpecfromFreq(), MassSpecProfile()
    '''

    def __init__(self, data_dict, d_params):

        """needs to simulate peak shape and pass as mz_exp and magnitude."""

        super().__init__([], [], d_params)

        self._set_parameters_objects(d_params)
        
        if self.label == Labels.thermo_centroid:
            self._baselise_noise = d_params.get("baselise_noise")
            self._baselise_noise_std = d_params.get("baselise_noise_std")

        self.is_centroid = True

        self.process_mass_spec(data_dict)

    def __simulate_profile__data__(self, exp_mz_centroid, magnitude_centroid):
        '''needs theoretical resolving power calculation and define peak shape
        this is a quick fix to trick a line plot be able to plot as sticks'''

        x, y = [], []
        for i in range(len(exp_mz_centroid)):
            x.append(exp_mz_centroid[i] - 0.0000001)
            x.append(exp_mz_centroid[i])
            x.append(exp_mz_centroid[i] + 0.0000001)
            y.append(0)
            y.append(magnitude_centroid[i])
            y.append(0)
        return x, y

    @property
    def mz_exp_profile(self):

        mz_list = []
        for mz in self.mz_exp:
            mz_list.append(mz - 0.0000001)
            mz_list.append(mz)
            mz_list.append(mz + 0.0000001)
        return mz_list
    
    @mz_exp_profile.setter
    def mz_exp_profile(self, _mz_exp ): self._mz_exp = _mz_exp

    @property
    def abundance_profile(self):
        ab_list = []
        for ab in self.abundance:
            ab_list.append(0)
            ab_list.append(ab)
            ab_list.append(0)
        return ab_list

    @abundance_profile.setter
    def abundance_profile(self, abundance ): self._abundance = abundance

    @property
    def tic(self):

        return sum(self.abundance)

    def process_mass_spec(self, data_dict, auto_noise=True, noise_bayes_est=False):
        import tqdm
        # overwrite process_mass_spec 
        # mspeak objs are usually added inside the PeaKPicking class 
        # for profile and freq based data
        

        s2n = True
        ion_charge = self.polarity
        #l_exp_mz_centroid = data_dict.get(Labels.mz)
        #l_intes_centr = data_dict.get(Labels.abundance)
        #l_peak_resolving_power = data_dict.get(Labels.rp)
        l_s2n = data_dict.get(Labels.s2n)
        
        if not l_s2n: s2n = False
        
        print("Loading mass spectrum object")
        
        for index, mz in enumerate(data_dict.get(Labels.mz)):
            
            # centroid peak does not have start and end peak index pos
            massspec_indexes = (index, index, index)
            
            if s2n:
                
                self.add_mspeak(
                    ion_charge,
                    mz,
                    data_dict.get(Labels.abundance)[index],
                    data_dict.get(Labels.rp)[index],
                    l_s2n[index],
                    massspec_indexes,
                    ms_parent=self
                )

            else:

                self.add_mspeak(
                    ion_charge,
                    mz,
                    data_dict.get(Labels.abundance)[index],
                    data_dict.get(Labels.rp)[index],
                    -999,
                    massspec_indexes,
                    ms_parent=self
                )

        self.mspeaks = self._mspeaks
        self._dynamic_range = self.max_abundance / self.min_abundance
        self._set_nominal_masses_start_final_indexes()
        
        if self.label != Labels.thermo_centroid:
            self._baselise_noise, self._baselise_noise_std = self.run_noise_threshold_calc(auto=auto_noise, bayes=noise_bayes_est)
            
class MassSpecCentroidLowRes(MassSpecCentroid,):
    
    '''Does not store MSPeak Objs, will iterate over mz, abundance pairs instead'''
    
    def __init__(self, data_dict, d_params):
    
        self._set_parameters_objects(d_params)
        self._mz_exp = array(data_dict.get(Labels.mz))
        self._abundance = array(data_dict.get(Labels.abundance))
        self._processed_tic = None
    
    def __len__(self):
        
        return len(self.mz_exp)
        
    def __getitem__(self, position):
        
        return (self.mz_exp[position], self.abundance[position])

    @property
    def mz_exp(self):
        
        return self._mz_exp 

    @property
    def abundance(self):

        return self._abundance

    @property
    def processed_tic(self):
    
        return sum(self._processed_tic)
    
    @property
    def tic(self):
    
        if self._processed_tic:
            return self._processed_tic
        else:
            return sum(self.abundance)
    
    @property
    def mz_abun_tuples(self):

        r = lambda x: ( int(round(x[0],0), int(round(x[1],0))) )

        return [r(i) for i in self]
    
    @property
    def mz_abun_dict(self):
        
        r = lambda x: int(round(x,0))
            
        return { r(i[0]):r(i[1]) for i in self}
    
    '''
    def to_excel(self, out_file_path):
    
    
    exportMS= HighResMassSpecExport(out_file_path, self)
    exportMS.to_excel()


    def to_hdf(self, out_file_path):
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS= HighResMassSpecExport(out_file_path, self)
        exportMS.to_hdf()

    def to_csv(self, out_file_path):
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS= HighResMassSpecExport(out_file_path, self)
        exportMS.to_csv()
        
    def to_pandas(self, out_file_path):
        #pickle dataframe (pkl extension)
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS= HighResMassSpecExport(out_file_path, self)
        exportMS.to_pandas()

    def to_dataframe(self,):
        #returns pandas dataframe
        
        from corems.mass_spectrum.output.export import HighResMassSpecExport
        exportMS= HighResMassSpecExport(self.filename, self)
        return exportMS.get_pandas_df()
       ''' 





__author__ = 'Yuri E. Corilo'
__date__ = 'Jul 02, 2019'

import dataclasses
import os
from typing import List, Dict

from corems.encapsulation.constant import Atoms, Labels

@dataclasses.dataclass
class TransientSetting:

    implemented_apodization_function: tuple = ('Hamming', 'Hanning', 'Blackman')
    apodization_method: str = 'Hanning'
    number_of_truncations: int = 0
    number_of_zero_fills: int = 1

    def __post_init__(self):

        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

@dataclasses.dataclass
class DataInputSetting:

    # add to this dict the VALUES to match your labels, THE ORDER WON'T MATTER
    # 'column_translate' : {'m/z':'m/z', 'Resolving Power':'Resolving Power', 'Abundance':'Abundance' , 'S/N':'S/N'}
    header_translate: dict = dataclasses.field(default_factory=dict)

    def __post_init__(self):

        self.header_translate = {'m/z': Labels.mz, 
                                 'mOz': Labels.mz,
                                 'Mass': Labels.mz,
                                 'Resolving Power': 'Resolving Power',
                                 'Res.': Labels.rp,
                                 'Intensity': Labels.abundance,
                                 'I': Labels.abundance,
                                 'Abundance': Labels.abundance,
                                 'Signal/Noise': Labels.s2n,
                                 'S/N': Labels.s2n,
                                 'abs_abu': Labels.abundance,
                                 'sn': Labels.s2n,
                                 'resolution': Labels.rp}

    def add_mz_label(self, label):

        self.header_translate[label] = Labels.mz

    def add_peak_height_label(self, label):

        self.header_translate[label] = Labels.abundance

    def add_sn_label(self, label):

        self.header_translate[label] = Labels.s2n

    def add_resolving_power_label(self, label):

        self.header_translate[label] = Labels.rp

@dataclasses.dataclass
class LiqChromatographSetting:
    start_scan: int = 1
    final_scan: int = 7

    def __post_init__(self):
        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

@dataclasses.dataclass
class MassSpectrumSetting:

    threshold_method: str = 'relative_abundance'

    implemented_noise_threshold_methods: tuple = ('auto', 'signal_noise', 'relative_abundance')

    noise_threshold_std: int = 6

    s2n_threshold: float = 4

    relative_abundance_threshold: float = 6  # from 0-100

    min_noise_mz: float = 100.0
    max_noise_mz: float = 1200.0

    min_picking_mz: float = 100.0
    max_picking_mz: float = 1200.0

    calib_minimize_method: str = 'Powell'
    calib_pol_order: int = 2
    max_calib_ppm_error: float = 1
    min_calib_ppm_error: float = -1
    calib_sn_threshold: float = 10

    do_calibration: bool = True

    def __post_init__(self):
        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)


@dataclasses.dataclass
class MassSpecPeakSetting:

    kendrick_base: Dict = dataclasses.field(default_factory=dict)

    # kendrick_base : Dict =  {'C': 1, 'H':2}

    peak_min_prominence_percent: float = 1  # 1-100 % used for peak detection

    peak_max_prominence_percent: float = 0.1  # 1-100 % used for baseline detection

    def __post_init__(self):

        # default to CH2
        if not self.kendrick_base:
            self.kendrick_base = {'C': 1, 'H': 2}
        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

@dataclasses.dataclass
class GasChromatographSetting:

    use_deconvolution: bool = False

    implemented_smooth_method: tuple = ('savgol', 'hanning', 'blackman', 'bartlett', 'flat', 'boxcar')

    smooth_window: int = 5

    smooth_method: str = 'savgol'

    savgol_pol_order: int = 2

    peak_height_max_percent: float = 10  # 1-100 % used for baseline detection use 0.1 for second_derivative and 10 for other methods

    peak_max_prominence_percent: float = 1  # 1-100 % used for baseline detection

    min_peak_datapoints: float = 5

    max_peak_width: float = 0.1

    noise_threshold_method: str = 'manual_relative_abundance'

    implemented_noise_threshold_methods: tuple = ('auto_relative_abundance', 'manual_relative_abundance', 'second_derivative')

    std_noise_threshold: int = 3

    peak_height_min_percent: float = 0.1  # 0-100 % used for peak detection

    peak_min_prominence_percent: float = 0.1  # 0-100 % used for peak detection

    eic_signal_threshold: float = 0.01  # 0-100 % used for extracted ion chromatogram peak detection

    max_rt_distance: float = 0.025  # minutes, max distance allowance hierarchical clutter

    def __post_init__(self):

        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

@dataclasses.dataclass
class CompoundSearchSettings:

    url_database: str = 'sqlite:///db/pnnl_lowres_gcms_compounds.sqlite'

    ri_search_range: float = 35

    rt_search_range: float = 1.0  # used for retention index calibration

    correlation_threshold: float = 0.5  # used for calibration, spectral similarity 

    score_threshold: float = 0.0

    ri_spacing: float = 200

    ri_std: float = 3  # in standard deviation

    ri_calibration_compound_names: List = dataclasses.field(default_factory=list)

    # calculates and export all spectral similarity methods
    exploratory_mode: bool = False

    score_methods: tuple = ('highest_sim_score', 'highest_ss')

    output_score_method: str = 'All'

    def __post_init__(self):
        # enforce datatype
        self.url_database = os.getenv('SPECTRAL_GCMS_DATABASE_URL', 'sqlite:///db/pnnl_lowres_gcms_compounds.sqlite')

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

        self.ri_calibration_compound_names = (' [C8] Methyl Caprylate [7.812]',
                                ' [C10] Methyl Caprate [10.647]',
                                ' [C9] Methyl Pelargonate [9.248]',
                                ' [C12] Methyl Laurate [13.250]',
                                ' [C14] Methyl Myristate [15.597]',
                                ' [C16] Methyl Palmitate [17.723]',
                                ' [C18] Methyl Stearate [19.663]',
                                ' [C20] Methyl Eicosanoate [21.441]',
                                ' [C22] Methyl Docosanoate [23.082]',
                                ' [C24] Methyl Linocerate [24.603]',
                                ' [C26] Methyl Hexacosanoate [26.023]',
                                ' [C28] Methyl Octacosanoate [27.349]',
                                ' [C30] Methyl Triacontanoate [28.72]')

@dataclasses.dataclass
class MolecularLookupDictSettings:

    '''
    ### DO NOT CHANGE IT! These are used to generate the database entries 

    ### DO change when creating a new application database 

    ### FOR search settings runtime and database query check use the MolecularFormulaSearchSettings class below

    ### C, H, N, O, S and P atoms are ALWAYS needed at usedAtoms
    ### if you don't want to include one of those atoms set the max and min at 0
    ### you can include any atom listed at Atoms class inside encapsulation.settings.constants module
    ### make sure to include the selected covalence at the used_atoms_valences when adding new atoms 
    ### NOTE : Adducts atoms have zero covalence
    ### NOTE : Not using static variable because this class is distributed using multiprocessing
    '''
    def __init__(self):

        self.usedAtoms = {'C': (1, 90),
                          'H': (4, 200),
                          'O': (0, 12),
                          'N': (0, 0),
                          'S': (0, 0),
                          'P': (0, 0),
                          'Cl': (0, 0),
                          }

        self.min_mz = 100

        self.max_mz = 1200

        self.min_dbe = 0

        self.max_dbe = 50

        # overwrites the dbe limits above to DBE = (C + heteroatoms) * 0.9
        self.use_pah_line_rule = False

        self.isRadical = True

        self.isProtonated = True

        self.min_op_filter = 2

        self.min_hc_filter = 0.3

        self.min_oc_filter = 1.2

        self.max_oc_filter = 0.0

        self.url_database = None

        self.db_jobs = 1

        self.used_atom_valences = {'C': 4,
                                   '13C': 4,
                                   'H': 1,
                                   'O': 2,
                                   '18O': 2,
                                   'N': 3,
                                   'S': 2,
                                   '34S': 2,
                                   'P': 3,
                                   'Cl': 1,
                                   '37Cl': 1,
                                   'Br': 1,
                                   'Na': 1,
                                   'F': 1,
                                   'K': 0,
                                   }

@dataclasses.dataclass
class MolecularFormulaSearchSettings:

    use_isotopologue_filter: bool = False

    isotopologue_filter_threshold: float = 33

    isotopologue_filter_atoms: tuple = ('Cl', 'Br')

    use_runtime_kendrick_filter: bool = False

    use_min_peaks_filter: bool = True

    min_peaks_per_class: int = 15

    url_database: str = 'postgresql+psycopg2://coremsappdb:coremsapppnnl@localhost:5432/coremsapp'

    db_jobs: int = 3

    '''query setting'''
    ion_charge: int = -1

    min_hc_filter: float = 0.3

    max_hc_filter: float = 3

    min_oc_filter: float = 0.0

    max_oc_filter: float = 1.2

    min_op_filter: float = 2

    use_pah_line_rule: bool = False

    min_dbe: float = 0

    max_dbe: float = 40

    mz_error_score_weight: float = 0.6

    isotopologue_score_weight: float = 0.4

    # look for close shell ions [M + Adduct]+ only considers metal set in the list adduct_atoms  
    adduct_atoms_neg: tuple = ('Cl', 'Br')

    adduct_atoms_pos: tuple = ('Na', 'K')

    score_methods: tuple = ('S_P_lowest_error', 'N_S_P_lowest_error', 'lowest_error', 'prob_score',
                            'air_filter_error', 'water_filter_error', 'earth_filter_error')

    score_method: str = 'prob_score'

    output_score_method: str = 'All Candidates'

    # depending on the polarity mode it looks for [M].+ , [M].-
    # query and automatically compile add entry if it doesn't exist

    isRadical: bool = False

    # depending on the polarity mode it looks for [M + H]+ , [M - H]+
    # query and automatically compile and push options if it doesn't exist
    isProtonated: bool = True

    isAdduct: bool = False

    usedAtoms: dict = dataclasses.field(default_factory=dict)

    ''' search setting '''

    ionization_type: str = 'ESI'

    # empirically set / needs optimization
    min_ppm_error: float = -10.0  # ppm

    # empirically set / needs optimization
    max_ppm_error: float = 10.0  # ppm

    # empirically set / needs optimization set for isotopologue search
    min_abun_error: float = -100.0  # percentage

    # empirically set / needs optimization set for isotopologue search
    max_abun_error: float = 100.0  # percentage

    # empirically set / needs optimization
    mz_error_range: float = 1.5

    # 'distance', 'lowest', 'symmetrical','average' 'None'
    error_method: str = 'None'

    mz_error_average: float = 0.0

    # used_atom_valences: {'C': 4, 'H':1, etc} = dataclasses.field(default_factory=dict)
    used_atom_valences: dict = dataclasses.field(default_factory=dict)

    def __post_init__(self):

        self.url_database = os.getenv('COREMS_DATABASE_URL', 'sqlite:///db/molformula.db')
        # enforce datatype
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):

                value = field.type(value)
                setattr(self, field.name, value)

        # enforce C and H if either do not exists
        if 'C' not in self.usedAtoms.keys():
            self.usedAtoms['C'] = (1, 100)
        if 'H' not in self.usedAtoms.keys():
            self.usedAtoms['H'] = (1, 200)

        # add cummon values
        current_used_atoms = self.used_atom_valences.keys()
        for atom in Atoms.atoms_covalence.keys():
            if atom not in current_used_atoms:
                self.used_atom_valences[atom] = Atoms.atoms_covalence.get(atom)

        '''
        self.used_atom_valences.update({'13C': 4,
                                        '18O': 2,
                                        '34S': 2,
                                        'Cl': 1,
                                        '37Cl': 1,
                                        'Br': 1,
                                        'Na': 1,
                                        'F': 1,
                                        'K': 1,
                                        })
        '''

if __name__ == '__main__':
    a = DataInputSetting()
    print(a.__dict__)



from comtypes import byref
from comtypes.automation import BSTR, VARIANT
from comtypes.client import CreateObject
from ctypes import c_double, c_long

from corems.encapsulation.factory.parameters import default_parameters
from corems.mass_spectra.factory.LC_Class import LCMSBase
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecProfile, MassSpecCentroid
from corems.encapsulation.constant import Labels

from threading import Thread
import multiprocessing
import numpy

__author__ = "Yuri E. Corilo"
__date__ = "July 9, 2019"

class ImportMassSpectraThermoMSFileReader(Thread):
    
    """     Read FULL and PROFILE (it ignores all other scans) raw file data and store it return a LCMS class
    *  Default behavior is to load all scans numbers

    *  set start_scan_number  and final_scan_number to change it before calling start(), or run()

    *  Noise threshold will break the mass_spec.process_mass_spec() if the method in the
    MassSpecSetting class is set to something other than Relative Abundance
    (it needs to be fixed to work with all methods)
    """

    def __init__(self, file_location):

        Thread.__init__(self)

        self.thermo_Library = CreateObject("MSFileReader.XRawfile")

        self.thermo_Library.open(file_location)

        self.res = self.thermo_Library.SetCurrentController(0, 1)

        self.check_load_success()

        self.LCMS = LCMSBase(file_location)

        self._initial_scan_number = 1

        self._final_scan_number = self.get_scans_numbers()

        self.file_location = file_location

    @property
    def initial_scan_number(self):
        return self._initial_scan_number

    @property
    def final_scan_number(self):
        return self._final_scan_number

    def check_scan(self, scan):

        scan_numbers = self.get_scans_numbers()
        return scan <= scan_numbers

    @initial_scan_number.setter
    def initial_scan_number(self, start_scan_number):
        if self.check_scan(start_scan_number):
            self._initial_scan_number = start_scan_number
        else:
            raise Exception(
                "startscan and finalscan should be less than %s"
                % self.get_scans_numbers()
            )

    @final_scan_number.setter
    def final_scan_number(self, final_scan_number):

        if self.check_scan(final_scan_number):
            self._final_scan_number = final_scan_number
        else:
            raise Exception(
                "startscan and finalscan should be less than %s"
                % self.get_scans_numbers()
            )

    
    def run(self):
        '''thread will automatically process mass spectrum
        use the get_mass_spectra class to import without processing mass spectrum'''

        d_parameters = default_parameters(self.file_location)
        self._import_mass_spectra(d_parameters)

        # return self.LCMS

    def get_mass_spectra(self,auto_process=True):

        d_parameters = default_parameters(self.file_location)
        self._import_mass_spectra(d_parameters, auto_process=auto_process)
        return self.LCMS

    def check_load_success(self):
        """ 0 if successful; otherwise, see Error Codes on MSFileReader Manual """
        if self.res == 0:

            self.break_it = False
            return True
        else:

            raise ImportError(str(self.res))

    def get_filter_for_scan_num(self, scan_number):
        """Returns the closest matching run time that corresponds to scan_number for the current
        controller. This function is only supported for MS device controllers.
        e.g.  ['FTMS', '-', 'p', 'NSI', 'Full', 'ms', '[200.00-1000.00]']
        """
        str_filter = BSTR(None)
        error = self.thermo_Library.GetFilterForScanNum(scan_number, byref(str_filter))
        if error:
            raise IOError(
                "scan %i GetFilterForScanNum error : %s" % (scan_number, str(error))
            )
        else:
            return str(str_filter.value).split()

    def check_full_scan(self, scan_number):

        scan_mode_symbol = self.get_filter_for_scan_num(scan_number)[4]

        return scan_mode_symbol == "Full"

    def get_polarity_mode(self, scan_number):

        polarity_symbol = self.get_filter_for_scan_num(scan_number)[1]

        if polarity_symbol == "+":

            return 1
            # return "POSITIVE_ION_MODE"

        elif polarity_symbol == "-":

            return -1

        else:

            raise Exception("Polarity Mode Unknown, please set it manually")

    def get_data(self, scan, d_parameter):

        scan = c_long(scan)
        pvarLabels = VARIANT()
        pvarFlags = VARIANT()

        self.thermo_Library.GetLabelData(pvarLabels, pvarFlags, scan)
        scans_labels = numpy.array(pvarLabels.value)

        mz = scans_labels[0]
        magnitude = scans_labels[1]
        rp = scans_labels[2]
        base_noise = scans_labels[3]
        noise = scans_labels[4]
        # charge = scans_labels[5]

        array_noise_std = (numpy.array(noise) - numpy.array(base_noise)) / 3
        l_signal_to_noise = numpy.array(magnitude) / array_noise_std

        d_parameter["baselise_noise"] = numpy.average(array_noise_std)

        d_parameter["baselise_noise_std"] = numpy.std(array_noise_std)

        data_dict = {
            Labels.mz: mz,
            Labels.abundance: magnitude,
            Labels.rp: rp,
            Labels.s2n: l_signal_to_noise,
        }

        return data_dict

    def get_scans_numbers(self):

        nScans = c_long()
        self.thermo_Library.GetNumSpectra(nScans)

        return int(nScans.value)

    def get_ScanHeaderInfoForScanNum(self, scan_number):

        nScanNumber = c_long(scan_number)  # get info for the twelfth scan
        nPackets = c_long(0)
        dRetantionTime = c_double(0.0)
        dLowMass = c_double(0.0)
        dHighMass = c_double(0.0)
        dTIC = c_double(0.0)
        dBasePeakMass = c_double(0.0)
        dBasePeakIntensity = c_double(0.0)
        nChannels = c_long(0)
        bUniformTime = c_long(False)
        dFrequency = c_double(0.0)
        self.thermo_Library.GetScanHeaderInfoForScanNum(
            nScanNumber,
            nPackets,
            dRetantionTime,
            dLowMass,
            dHighMass,
            dTIC,
            dBasePeakMass,
            dBasePeakIntensity,
            nChannels,
            bUniformTime,
            dFrequency,
        )

        return dRetantionTime.value, dTIC.value

    def is_profile_scan_for_scan_num(self, scan_number):

        IsProfileScan = c_long()
        error = self.thermo_Library.IsProfileScanForScanNum(
            c_long(scan_number), byref(IsProfileScan)
        )
        if error:
            raise IOError("IsProfileScanForScanNum error :", error)
        # print (IsProfileScan.value, bool(1))
        return bool(IsProfileScan.value)

    def _import_mass_spectra(self, d_params, auto_process=True):
        results = []
        # Each_Mass_Spectrum = namedtuple('each_mass_spectrum', ['mass_list', 'abundance_list', 'retention_time', 'scan_number', 'tic_number'])

        if self.check_load_success():

            """get number of scans"""

            list_Tics = list()

            list_RetentionTimeSeconds = list()

            list_scans = list()

            """key = scan_number or retention time"""
            # print(self.initial_scan_number, self.final_scan_number)
            for scan_number in range(
                self.initial_scan_number, self.final_scan_number + 1
            ):
                #print(scan_number)
                # scan_number = scan_number + 1

                "only import FULL scans and Profile Mode, it ignores all others"

                if self.check_full_scan(scan_number):

                    if self.is_profile_scan_for_scan_num(scan_number):

                        d_params["label"] = Labels.thermo_centroid

                        d_params["polarity"] = self.get_polarity_mode(scan_number)

                        d_params["rt"], TIC = self.get_ScanHeaderInfoForScanNum(
                            scan_number
                        )

                        d_params["scan_number"] = scan_number

                        list_RetentionTimeSeconds.append(d_params.get("rt"))

                        list_Tics.append(TIC)

                        list_scans.append(scan_number)

                        data_dict = self.get_data(scan_number, d_params)

                        #results.append((data, d_params))
                        
                        mass_spec = MassSpecCentroid(data_dict, d_params)
                        
                        self.LCMS.add_mass_spectrum(mass_spec)

            #pool = multiprocessing.Pool(5)
            #result = pool.starmap(MassSpecCentroid, results)
            #for ms in result:
            #self.LCMS.add_mass_spectrum(ms)
            
            self.LCMS.retention_time = list_RetentionTimeSeconds
            self.LCMS.set_tic_list(list_Tics)
            self.LCMS.set_scans_number_list(list_scans)

    def get_lcms(self):
        """get_lc_ms_class method should only be used when using this class as a Thread, 
        otherwise use the run() method to return the LCMS class"""

        if self.LCMS.get(self._initial_scan_number):
            return self.LCMS
        else:
            raise Exception("returning a empty LCMS class")



from corems.molecular_id.calc.ClusterFilter import ClusteringFilter

class MolecularFormulaSearchFilters:

    @staticmethod
    def filter_kendrick( ms_peak_indexes, mass_spectrum_obj):

        index_to_remove = []
        
        if mass_spectrum_obj.molecular_search_settings.use_runtime_kendrick_filter:
            
            index_to_remove = ClusteringFilter().filter_kendrick_by_index(ms_peak_indexes, mass_spectrum_obj)

            #for index in noise_indexes: self.mass_spectrum_obj[index].clear_molecular_formulas()
        
        all_index_to_remove = []

        for peak_index, mf_obj in index_to_remove:
                
                ms_peak_indexes.remove((peak_index, mf_obj))

                all_index_to_remove.extend(mf_obj.mspeak_mf_isotopologues_indexes)

        all_index_to_remove =  list(set(all_index_to_remove + index_to_remove))

        for peak_index, mf_obj in all_index_to_remove:
            
            mass_spectrum_obj[peak_index].remove_molecular_formula(mf_obj)
            
        return ms_peak_indexes

    @staticmethod
    def check_min_peaks( ms_peak_indexes, mass_spectrum):
        
        if mass_spectrum.molecular_search_settings.use_min_peaks_filter:

            if not len(ms_peak_indexes) >= mass_spectrum.molecular_search_settings.min_peaks_per_class:
                
                for peak_index, mf_obj in ms_peak_indexes:
                
                    mass_spectrum[peak_index].remove_molecular_formula(mf_obj)

    @staticmethod
    def filter_isotopologue( ms_peak_indexes, mass_spectrum):
        
        index_to_remove = []
        #print(len(ms_peak_indexes))
        if mass_spectrum.molecular_search_settings.use_isotopologue_filter:

            atoms_iso_filter = mass_spectrum.molecular_search_settings.isotopologue_filter_atoms

            isotopologue_count_threshold = mass_spectrum.molecular_search_settings.isotopologue_filter_threshold

            for mspeak_index, mf_obj in ms_peak_indexes:
            
                if mf_obj.isotopologue_count_percentile < isotopologue_count_threshold:
                    
                    if set(mf_obj.atoms).intersection(atoms_iso_filter):
                        
                        #removes tuple obj from initial list to be used on next filter steps
                        ms_peak_indexes.remove((mspeak_index, mf_obj))
                        
                        # current mf_obj
                        index_to_remove.append((mspeak_index, mf_obj))
                        # all other associated isotopolgues
                        index_to_remove.extend(mf_obj.mspeak_mf_isotopologues_indexes)

        #iterate over all indexes to be remove and remove the mf from the mspeak 
        
        #print(len(ms_peak_indexes))
        for peak_index, mf_obj in index_to_remove:
                #print(peak_index, mf_obj)
                mass_spectrum[peak_index].remove_molecular_formula(mf_obj)

        
        return ms_peak_indexes 


import numpy as np

class FreqDomain_Calibration:

    def __init__(self, mass_spectrum, selected_mass_peaks, include_isotopologue=False):

        self.selected_mspeaks = selected_mass_peaks
        error = list()
        freq_exp = list()
        mz_calc = list()
        mz_exp = list()

        for mspeak in selected_mass_peaks:

            if not include_isotopologue:
                molecular_formulas = [
                    formula for formula in mspeak if not formula.is_isotopologue]
            else:
                molecular_formulas = mspeak

            for molecular_formula in molecular_formulas:

                freq_exp.append(mspeak.freq_exp)
                error.append(
                    molecular_formula.mz_error)
                mz_calc.append(molecular_formula.mz_calc)
                mz_exp.append(mspeak.mz_exp)

        self.mz_exp = np.array(mz_exp)
        self.mz_calc = np.array(mz_calc)
        self.freq_exp = np.array(freq_exp)
        self.mass_spectrum = mass_spectrum
        self.freq_exp_ms = np.array(
            [mspeak.freq_exp for mspeak in mass_spectrum])

    def recal_mass_spec(self, mz_domain, Aterm, Bterm, Cterm):

        self.mass_spectrum._calibration_terms = (Aterm, Bterm, 0)
        
        #for indexes, mspeak in enumerate(self.mass_spectrum):
        #    mspeak.mz_cal = mz_domain[indexes] 
        self.mass_spectrum.mz_cal = mz_domain

    def linear(self):

        matrix = np.vstack([1/self.freq_exp, np.ones(len(self.freq_exp))]).T
        Aterm, Bterm = np.linalg.lstsq(matrix, self.mz_calc, rcond=None)[0]
        print("%.2f Aterm,  %.2f Bterm" %  (Aterm, Bterm))
        print('Linear Calibration %.2f Aterm,  %.2f Bterm ' %(Aterm, Bterm))
        print()
        #mz_domain = Aterm / (self.freq_exp_ms + Bterm)
        mz_domain = (Aterm/self.freq_exp_ms) + Bterm
        self.recal_mass_spec(mz_domain, Aterm, Bterm, 0)

    def quadratic(self, iteration=False):

        mz_calc = self.mz_calc
        freq_exp = self.freq_exp
        mz_exp = self.mz_exp

        error = ((mz_exp-mz_calc)/mz_calc) * 1000000
        last_rms = np.sqrt(np.mean(error**2))
        while True:

            matrix = np.vstack(
                [1/freq_exp, 1/np.power(freq_exp, 2), np.ones(len(freq_exp))]).T
            Aterm, Bterm, Cterm = np.linalg.lstsq(
                matrix, self.mz_calc, rcond=None)[0]
            mz_exp = (Aterm / (freq_exp)) + \
                (Bterm / np.power((freq_exp), 2)) + Cterm
            error = ((mz_exp-mz_calc)/mz_calc)*1000000
            rms = np.sqrt(np.mean(error**2))
            std = np.std(error)
            print("%.2f Aterm,  %.2f Bterm" %  (Aterm, Bterm))
            print('Quadratic Calibration %.2f RMS,  %.2f std,  %.2f Aterm,  %.2f Bterm ' %(rms, std, Aterm, Bterm))
            print()
            if rms < last_rms:
                last_rms = rms
                freq_exp = (Aterm + np.sqrt(np.power(-Aterm, 2) -
                                            (4*Cterm*(mz_exp-Bterm)))) / (2*mz_exp)

                mz_domain = (Aterm / (self.freq_exp_ms)) + \
                    (Bterm / np.power((self.freq_exp_ms), 2)) + Cterm
                self.recal_mass_spec(mz_domain, Aterm, Bterm, Cterm)
                if not iteration:
                    break
            else:
                break

    def ledford_calibration(self, iteration=False):

        mz_calc = self.mz_calc
        freq_exp = self.freq_exp
        mz_exp = self.mz_exp

        error = ((mz_exp-self.mz_calc)/self.mz_calc) * 1000000
        last_rms = np.sqrt(np.mean(error**2))
        while True:

            matrix = np.vstack([1/freq_exp, 1/np.power(freq_exp, 2)]).T
            Aterm, Bterm = np.linalg.lstsq(matrix, self.mz_calc, rcond=None)[0]

            mz_exp = (Aterm / (freq_exp)) + (Bterm / np.power((freq_exp), 2))
            error = ((mz_exp-mz_calc)/mz_calc)*1000000
            rms = np.sqrt(np.mean(error**2))
            std = np.std(error)
            print("%.2f Aterm,  %.2f Bterm" %  (Aterm, Bterm))
            print('Ledford Calibration %.2f RMS,  %.2f std,  %.2f Aterm,  %.2f Bterm ' %(rms, std, Aterm, Bterm))
            print()
            if rms < last_rms:
                last_rms = rms
                freq_exp = (Aterm + np.sqrt(np.power(-Aterm, 2) -
                                            (4*mz_exp-Bterm))) / (2*mz_exp)
                mz_domain = (Aterm / (self.freq_exp_ms)) + \
                    (Bterm / np.power((self.freq_exp_ms), 2))
                self.recal_mass_spec(mz_domain, Aterm, Bterm, 0)
                if not iteration:
                    break
            else:
                break

    def step_fit(self, steps=4):

        def f_to_mz(f, A, B, C, a): 
                return (A / f) + (B / np.power(f, 2)) + (C*a / np.power(f, 2))
        
        def mz_to_f(m, A, B, C): return (-A-m/B)
        
        tuple_indexes = [(i, i+steps) for i in range(0, len(self.selected_mspeaks)-steps, steps)]

        for current_index, tuple_index in enumerate(tuple_indexes):
            
            mspeak_ii, mspeak_fi = tuple_index
            freq_exp = list()
            mz_calc = list()
            mz_exp = list()
            abu = list()
            
            for i in range(mspeak_ii, mspeak_fi+1):

                best_formula = self.selected_mspeaks[i].best_molecular_formula_candidate

                freq_exp.append(self.selected_mspeaks[i].freq_exp)
                mz_calc.append(best_formula.mz_calc)
                mz_exp.append(self.selected_mspeaks[i].mz_exp)
                abu.append(self.selected_mspeaks[i].abundance)
                        
            
            freq_exp = np.array(freq_exp)
            mz_calc = np.array(mz_calc)
            mz_exp = np.array(mz_exp)
            abu = np.array(abu)

            if current_index == len(tuple_indexes)-1:
                ms_peaks_indexes = (self.selected_mspeaks[mspeak_ii].index, 0)
            
            elif current_index == 0:
                ms_peaks_indexes = (len(self.mass_spectrum)-1,
                                    self.selected_mspeaks[mspeak_fi].index-1)
            else:
                ms_peaks_indexes = (
                    self.selected_mspeaks[mspeak_ii].index, self.selected_mspeaks[mspeak_fi].index-1)

            final_index, start_index = ms_peaks_indexes
                           
            matrix = np.vstack([1/freq_exp, 1/np.power(freq_exp, 2)]).T
            A, B = np.linalg.lstsq(matrix, mz_calc, rcond=None)[0]
            C = 0
            
            for mspeak in self.mass_spectrum[start_index:final_index]:
                mspeak.mz_cal = f_to_mz(mspeak.freq_exp, A, B, C, 0)
        
        self.mass_spectrum.is_calibrated = True    




__author__ = "Yuri E. Corilo"
__date__ = "Jun 12, 2019"
from numpy import array, trapz    
from corems.chroma_peak.calc.ChromaPeakCalc import GCPeakCalculation 
from corems.molecular_id.factory.EI_SQL import LowResCompoundRef

class ChromaPeakBase():
    '''
    classdocs
    '''
    def __init__(self, chromatogram_parent, mass_spectrum_obj, start_index, index, final_index):
        
        self.start_index = start_index
        self.final_index = final_index
        self.index = int(index)
        self.chromatogram_parent = chromatogram_parent
        self.mass_spectrum = mass_spectrum_obj
        'updated individual calculation'
        self._area = None
       
        self._compounds = []
   
    def __len__(self):
        
        return len(self._compounds)
        
    def __getitem__(self, position):
        
        return self._compounds[position]

    def add_compound(self, compounds_dict, similarity):
        #implemented in child class
        pass

    def remove_compound(self, compounds_obj):
        
        self._compounds.remove(compounds_obj)
        
    def clear_compounds(self):
        
        self._compounds = []

    @property   
    def rt(self):
        return self.mass_spectrum.rt

    @property   
    def tic(self):
        return self.mass_spectrum.tic    

    @property   
    def area(self):
        return self._area

    @property
    def rt_list(self):
        return [self.chromatogram_parent.retention_time[i] for i in range(self.start_index, self.final_index+1) ]
    
    @property   
    def tic_list(self):
        return [self.chromatogram_parent.tic[i] for i in range(self.start_index, self.final_index+1) ]

   
class GCPeak(ChromaPeakBase, GCPeakCalculation):

    def __init__(self, chromatogram_parent, mass_spectrum_obj, indexes):
    
        super().__init__(chromatogram_parent, mass_spectrum_obj, *indexes)
        
        self._ri = None

    def add_compound(self, compounds_dict, spectral_similarity_scores, ri_score=None, similarity_score=None):
       
        compound_obj = LowResCompoundRef(compounds_dict)

        #add all spectral similarities methods as a dict
        compound_obj.spectral_similarity_scores = spectral_similarity_scores
        #TODO need to add spectral similarity score label in the options in the parameters encapsulation class
        compound_obj.spectral_similarity_score = spectral_similarity_scores.get("cosine_correlation")

        compound_obj.ri_score = ri_score

        compound_obj.similarity_score = similarity_score
        
        self._compounds.append(compound_obj)
        
        if similarity_score:
            self._compounds.sort(key=lambda c: c.similarity_score, reverse=True)
        else:
            self._compounds.sort(key=lambda c: c.spectral_similarity_score, reverse=True)
    
    @property
    def ri(self): return self._ri

    @property
    def highest_ss_compound(self):
        if self:
            return max(self, key = lambda c: c.spectral_similarity_score)
        else:
            None
    
    @property
    def highest_score_compound(self):
        if self:
            return max(self, key = lambda c: c.similarity_score)
        else: 
            return None

    @property
    def compound_names(self):
        if self:
            return [c.name for c in self]
        else: 
            return []

class GCPeakDeconvolved(GCPeak):
    
    def __init__(self, chromatogram_parent, mass_spectra, apex_index, rt_list, tic_list ):
        
        self._ri = None   

        self._rt_list = list(rt_list)
        
        self._tic_list = list(tic_list)

        self.mass_spectra = list(mass_spectra)

        super().__init__(chromatogram_parent, self.mass_spectra[apex_index], (0, apex_index, len(self.mass_spectra)-1))

    @property
    def rt_list(self):
        return self._rt_list
    
    @property   
    def tic_list(self):   
        return self._tic_list



import os,  sys
sys.path.append('.')
from copy import deepcopy
from threading import Thread
from itertools import product

import tqdm

from corems.encapsulation.constant import Labels, Atoms
from corems.molecular_id.calc.MolecularFilter import MolecularFormulaSearchFilters
from corems.molecular_id.factory.MolecularLookupTable import MolecularCombinations
from corems.molecular_id.search.findOxygenPeaks import FindOxygenPeaks
from corems.molecular_id.search.molecularFormulaSearch import SearchMolecularFormulaWorker
from corems.molecular_id.factory.molecularSQL import MolForm_SQL
from corems.molecular_id.calc.ClusterFilter import ClusteringFilter
import json


class OxygenPriorityAssignment(Thread):

    def __init__(self, mass_spectrum_obj, sql_db=False):
        '''TODO:- add support for other atoms and adducts: Done
                - add dbe range on search runtime : Done
                - add docs
                - improve performace : Done 
        '''
        Thread.__init__(self)
        self.mass_spectrum_obj = mass_spectrum_obj
        #  initiated at create_molecular_database()
        #self.dict_molecular_lookup_table = None
        
        if not sql_db:

            self.sql_db = MolForm_SQL(url=mass_spectrum_obj.molecular_search_settings.url_database)
        
        else:
            
            self.sql_db = sql_db

    def run(self):
        
        # get Oxygen classes dict and the associate mspeak class 
        # list_of_classes_min_max_dbe = self.class_and_dbes_in_order()
        # create database separated to give the user the chance to use mass spec filters
             
        assign_classes_order_str_dict_tuple_list = self.create_data_base()
        
        if assign_classes_order_str_dict_tuple_list:
            
            self.run_worker_mass_spectrum(assign_classes_order_str_dict_tuple_list)
        
        else:
            
            raise RuntimeError('call create_data_base() first')

        self.sql_db.close()   

    def create_data_base(self):
        
        def create_molecular_database():
            
            # checks and creates the database entries for the specified heteroatomic classes 
            
            min_o = min(self.mass_spectrum_obj, key=lambda msp: msp[0]['O'])[0]['O'] - 2
            
            if min_o <= 0:
                min_o = 1

            max_o = max(self.mass_spectrum_obj, key=lambda msp: msp[0]['O'])[0]['O'] + 2

            #min_dbe = min(self.mass_spectrum_obj, key=lambda msp: msp[0].dbe)[0].dbe

            #max_dbe = max(self.mass_spectrum_obj, key=lambda msp: msp[0].dbe)[0].dbe

            #self.lookupTableSettings.use_pah_line_rule = False
            
            #self.lookupTableSettings.min_dbe = min_dbe/2#min_dbe - 7 if  (min_dbe - 7) > 0 else 0
            
            #self.lookupTableSettings.max_dbe = max_dbe * 2 #max_dbe + 7
            
            self.mass_spectrum_obj.reset_indexes()

            self.mass_spectrum_obj.filter_by_noise_threshold()
            
            #initial_ox = deepcopy(self.mass_spectrum_obj.molecular_search_settings.usedAtoms)

            self.mass_spectrum_obj.molecular_search_settings.usedAtoms['O'] = (min_o, max_o)

            self.nominal_mzs = self.mass_spectrum_obj.nominal_mz

        
        # get the most abundant peak and them every 14Da, only allow Ox and its derivatives
        print("Getting Oxygen Series")
        find_formula_thread = FindOxygenPeaks(self.mass_spectrum_obj, self.sql_db)
        find_formula_thread.run()
        
        #mass spec obj indexes are set to interate over only the peaks with a molecular formula candidate
        print("Getting Oxygen Series")
        find_formula_thread.set_mass_spec_indexes_by_found_peaks()
        
        #get the Ox class and the DBE for the lowest error molecular formula candidate
        dict_ox_class_and_ms_peak = self.ox_classes_and_peaks_in_order_()
                      
        # sort the classes by abundance
        print("Getting Oxygen Series Order")
        assign_classes_order_str_dict_tuple_list = self.get_classes_in_order(dict_ox_class_and_ms_peak)
        
        create_molecular_database()
                
        return assign_classes_order_str_dict_tuple_list
        
    def run_worker_mass_spectrum(self, assign_classes_order_tuples):
        
        def check_adduct_class(classe_dict):

            return any([key in classe_dict.keys() for key in self.mass_spectrum_obj.molecular_search_settings.adduct_atoms_neg])
        
        def set_min_max_dbe_by_oxygen(classe_dict):
            # calculates min and max DBE based on the Oxygen number
            # ref :https://pubs.acs.org/doi/full/10.1021/ac200464q
            # if class does not has O it use the pha rule
            # ref : Vlad Lobodin manuscript to be include here
            '''
            atoms_exchanges = ['N']
            if 'O' in classe_dict.keys():
                
                Oxygen_number = classe_dict.get("O")
                for atom in atoms_exchanges:
                    if atom in classe_dict.keys():
                        Oxygen_number += classe_dict.get(atom)

                self.mass_spectrum_obj.molecular_search_settings.min_dbe = (Oxygen_number/3) - 0.5 
                self.mass_spectrum_obj.molecular_search_settings.max_dbe = Oxygen_number*3 + 0.5 + 2
            
            else:
            '''    
            self.mass_spectrum_obj.molecular_search_settings.use_pah_line_rule = True

        def run_search(possible_formulas_dict, mass_spectrum_obj, min_abundance):
            
            all_assigned_indexes = list()
            
            for ms_peak in mass_spectrum_obj.sort_by_abundance():

                if ms_peak: continue
                #already assigned a molecular formula
               
                nominal_mz  = ms_peak.nominal_mz_exp

                #get mono isotopic peaks that was added a molecular formula obj
                #TODO update error variables

                possible_formulas_nominal = possible_formulas_dict.get(nominal_mz)
                
                if possible_formulas_nominal:

                    ms_peak_indexes = SearchMolecularFormulaWorker().find_formulas(possible_formulas_nominal, min_abundance, mass_spectrum_obj, ms_peak)    

                    all_assigned_indexes.extend(ms_peak_indexes)
            
            
            #filter peaks by percentile threshold of found isotopologues 
            all_assigned_indexes = MolecularFormulaSearchFilters().filter_isotopologue(all_assigned_indexes, mass_spectrum_obj)

            #filter noise by kendrick density
            all_assigned_indexes = MolecularFormulaSearchFilters().filter_kendrick(all_assigned_indexes, mass_spectrum_obj)

            #filter per min peaks per mono isotopic class
            # this function should always be the last filter, 
            # thefore no need to return remaining indexes
            MolecularFormulaSearchFilters().check_min_peaks(all_assigned_indexes, mass_spectrum_obj)

        #error_average = self.mass_spectrum_obj.molecular_search_settings.mz_error_average
        
        kdm_base = self.mass_spectrum_obj.mspeaks_settings.kendrick_base
        
        self.mass_spectrum_obj.change_kendrick_base_all_mspeaks(kdm_base)

        ClusteringFilter().filter_kendrick(self.mass_spectrum_obj)

        min_abundance = self.mass_spectrum_obj.min_abundance

        list_classes_str = [i[0] for i in assign_classes_order_tuples]

        pbar = tqdm.tqdm(assign_classes_order_tuples)
        
        dict_molecular_lookup_table = self.get_dict_molecular_database(list_classes_str)

        for classe_tuple in pbar:

            classe_str  = classe_tuple[0]
            classe_dict = classe_tuple[1]
            
            set_min_max_dbe_by_oxygen(classe_dict)
            
            #if len(classe_dict.keys()) == 2:
            #    if classe_dict.get('S') == 1:
            #       continue
            # limits the dbe by the Ox class most abundant,
            # need to add other atoms contribution to be more accurate
            # but +-7 should be sufficient to cover the range 
            
            if self.mass_spectrum_obj.molecular_search_settings.isProtonated:

                    #tqdm.set_description_str(desc=None, refresh=True)
                    pbar.set_description_str(desc="Started molecular formula search for class %s, (de)protonated " % classe_str, refresh=True)

                    ion_type = Labels.protonated_de_ion

                    possible_formulas_dict = dict_molecular_lookup_table.get(ion_type).get(classe_str)
                    
                    if possible_formulas_dict:

                        run_search(possible_formulas_dict, self.mass_spectrum_obj, min_abundance)

            if self.mass_spectrum_obj.molecular_search_settings.isRadical:

                    #print("Started molecular formula search for class %s,  radical" % classe_str)
                    pbar.set_description_str(desc="Started molecular formula search for class %s, radical" % classe_str, refresh=True)

                    ion_type = Labels.radical_ion
                    
                    possible_formulas_dict = dict_molecular_lookup_table.get(ion_type).get(classe_str)
                    
                    if possible_formulas_dict:

                        run_search(possible_formulas_dict, self.mass_spectrum_obj, min_abundance)

            # looks for adduct, used_atom_valences should be 0 
            # this code does not support H exchance by halogen atoms
            if self.mass_spectrum_obj.molecular_search_settings.isAdduct:
                
                pbar.set_description_str(desc="Started molecular formula search for class %s, adduct" % classe_str, refresh=True)
                #print("Started molecular formula search for class %s, adduct" % classe_str)
                
                ion_type = Labels.radical_ion
                
                possible_formulas_dict = dict_molecular_lookup_table.get(ion_type).get(classe_str)

                possible_formulas_adduct =self.add_adducts(possible_formulas_dict)

                if possible_formulas_adduct:

                    run_search(possible_formulas_adduct, self.mass_spectrum_obj, min_abundance)
        
        
    def get_dict_molecular_database(self, classe_str_list):
        
        nominal_mzs = self.nominal_mzs
        mf_search_settings = self.mass_spectrum_obj.molecular_search_settings
        ion_charge = self.mass_spectrum_obj.polarity

        sql_db = MolForm_SQL(url=mf_search_settings.url_database)
        
        dict_res = {}

        if mf_search_settings.isProtonated:
            dict_res[Labels.protonated_de_ion] = sql_db.get_dict_by_classes(classe_str_list, Labels.protonated_de_ion, nominal_mzs, ion_charge, mf_search_settings)    
            
        if mf_search_settings.isRadical:
            dict_res[Labels.radical_ion] = sql_db.get_dict_by_classes(classe_str_list, Labels.radical_ion, nominal_mzs, ion_charge, mf_search_settings)    

        if mf_search_settings.isAdduct:
            
            adduct_list = mf_search_settings.adduct_atoms_neg if ion_charge < 0 else mf_search_settings.adduct_atoms_pos
            dict_res[Labels.adduct_ion] = sql_db.get_dict_by_classes(classe_str_list, Labels.adduct_ion, nominal_mzs, ion_charge, mf_search_settings, adducts=adduct_list)    

        return dict_res

    def ox_classes_and_peaks_in_order_(self) -> dict:
        # order is only valid in python 3.4 and above
        # change to OrderedDict if your version is lower
        dict_ox_class_and_ms_peak = dict()
        
        for mspeak in self.mass_spectrum_obj.sort_by_abundance(reverse=True):
            
            #change this filter to cia filter, give more option here, confidence, number of isotopologue found etc

            ox_classe = mspeak.best_molecular_formula_candidate.class_label
            
            if ox_classe in dict_ox_class_and_ms_peak.keys():
                
                #get the most abundant of the same ox class
                if mspeak.abundance > dict_ox_class_and_ms_peak[ox_classe].abundance:

                    dict_ox_class_and_ms_peak[ox_classe] = (mspeak)
            else:
                    
                dict_ox_class_and_ms_peak[ox_classe] = (mspeak)
        
        return dict_ox_class_and_ms_peak

    def get_classes_in_order(self, dict_ox_class_and_ms_peak)-> [(str, dict)]: 
        ''' structure is 
            ('HC', {'HC': 1})'''
        
        usedAtoms = deepcopy(self.mass_spectrum_obj.molecular_search_settings.usedAtoms)
        
        usedAtoms.pop("C")
        usedAtoms.pop("H")
        usedAtoms.pop("O")

        min_n, max_n = usedAtoms.get('N') if usedAtoms.get('N') else (0,0)
        min_s, max_s = usedAtoms.get('S') if usedAtoms.get('S') else (0,0)
        min_p, max_p = usedAtoms.get('P') if usedAtoms.get('P') else (0,0)

        possible_n = [n for n in range(min_n, max_n + 1)]
        possible_s = [s for s in range(min_s, max_s + 1)]
        possible_p = [p for p in range(min_p, max_p + 1)]
        
        #used to enforce order for commum atoms 
        # and track the atom index in on the tuple in all_atoms_tuples
        atoms_in_order = ['N', 'S', 'P']
        
        #do number atoms prodcut and remove then from the usedAtoms dict
        all_atoms_tuples = product(possible_n, possible_s, possible_p)
        for atom in atoms_in_order:
            
            usedAtoms.pop(atom, None)
        
        #iterate over other atoms besides C,H, N, O, S and P
        
        for selected_atom_label, min_max_tuple in usedAtoms.items():
            
            min_x = min_max_tuple[0]
            max_x = min_max_tuple[1]

            possible_x = [x for x in range(min_x, max_x + 1)]
            all_atoms_tuples = product(all_atoms_tuples, possible_x)
            
            #merge tuples
            all_atoms_tuples = [all_atoms_combined[0] + (all_atoms_combined[1],) for all_atoms_combined in
                                all_atoms_tuples]
            
            #add atom label to the atoms_in_order list
            
            #important to index where the atom position is in on the tuple in all_atoms_tuples
            atoms_in_order.append(selected_atom_label)

        classes_strings_dict_tuples, hc_class = self.get_class_strings_dict(all_atoms_tuples, atoms_in_order)

        combined_classes = self.combine_ox_class_with_other(atoms_in_order, classes_strings_dict_tuples, dict_ox_class_and_ms_peak)
        
        combination_classes_ordered = self.sort_classes(atoms_in_order, combined_classes)
        
        oxygen_class_str_dict_tuple = [(ox_class, mspeak[0].class_dict) for ox_class, mspeak in dict_ox_class_and_ms_peak.items()] 

        ## add classes together and ignores classes selected from the main series
        for class_tuple in  combination_classes_ordered:
            if class_tuple not in oxygen_class_str_dict_tuple:
                oxygen_class_str_dict_tuple.append(class_tuple)
        
        return oxygen_class_str_dict_tuple

    @staticmethod
    def get_class_strings_dict(all_atoms_tuples, atoms_in_order) -> [(str, dict)]: 
        
        classe_list= []
        hc_class = []
        
        for all_atoms_tuple in all_atoms_tuples:
            
            classe_str = ''
            classe_dict = dict()
            
            for each_atoms_index, atoms_number in enumerate(all_atoms_tuple):
                
                if atoms_number != 0:
                    
                    classe_str = (classe_str + atoms_in_order[each_atoms_index] + str(atoms_number) + ' ')
                    
                    classe_dict[atoms_in_order[each_atoms_index]] = atoms_number

            classe_str = classe_str.strip()
            
            if len(classe_str) > 0:
            
                classe_list.append((classe_str,classe_dict))

            elif len(classe_str) == 0:

                hc_class.append(('HC', {'HC':1}))
        
        return classe_list, hc_class
    
    @staticmethod
    def combine_ox_class_with_other( atoms_in_order, classes_strings_dict_tuples, dict_ox_class_and_ms_peak) -> [dict]:
        
        #sort methods that uses the key of classes dictionary and the atoms_in_order as reference
        # c_tuple[1] = class_dict, because is one key:value map we loop through keys and get the first item only 
        # sort by len first then sort based on the atoms_in_order list
        atoms_in_order = Atoms.atoms_order

        Oxygen_mfs = dict_ox_class_and_ms_peak.values()
        
        
        #sort_method = lambda word: (len(word[0]), [atoms_in_order.index(atom) for atom in list( word[1].keys())])
        
        #print(classes_strings_dict_tuples)
        #classe_in_order = sorted(classes_strings_dict_tuples, key = sort_method)
        #print(classe_in_order)
        
        combination = []
        
        # _ ignoring the class_str
        for _ , other_classe_dict in classes_strings_dict_tuples:
          
           #combination.extend([[other_classe_str + ' ' + Oxygen_mf[0].class_label , {**other_classe_dict, **Oxygen_mf[0].class_dict}] for Oxygen_mf in Oxygen_mfs])
           combination.extend([{**other_classe_dict, **Oxygen_mf[0].class_dict} for Oxygen_mf in Oxygen_mfs])
 
        return combination
    
    @staticmethod
    def sort_classes( atoms_in_order, combination_tuples) -> [(str, dict)]: 
        
        join_list_of_list_classes = list()
        atoms_in_order =  ['N','S','P','O'] + atoms_in_order[3:]
        
        sort_method = lambda atoms_keys: [atoms_in_order.index(atoms_keys)] #(len(word[0]), print(word[1]))#[atoms_in_order.index(atom) for atom in list( word[1].keys())])
        for class_dict in combination_tuples:
            
            sorted_dict_keys = sorted(class_dict, key = sort_method)
            class_str = ' '.join([atom + str(class_dict[atom]) for atom in sorted_dict_keys])
            new_class_dict = { atom: class_dict[atom] for atom in sorted_dict_keys}
            join_list_of_list_classes.append((class_str, new_class_dict))
        
        return join_list_of_list_classes
 
    def add_adducts(self, possible_formulas):
        
        ion_type = Labels.adduct_ion

        if self.mass_spectrum_obj.polarity < 0:
            adduct_atoms = self.mass_spectrum_obj.molecular_search_settings.adduct_atoms_neg
            molform_model = MolecularFormulaTableNeg
        else:
            adduct_atoms = self.mass_spectrum_obj.molecular_search_settings.adduct_atoms_pos
            molform_model = MolecularFormulaTablePos

        new_dict = {}
        
        for nominal_mz, list_formulas in possible_formulas.items():
            
            for adduct_atom in adduct_atoms:
                
                adduct_atom_mass= Atoms.atomic_masses.get(adduct_atom) 

                for molecularFormulaTable in  list_formulas:
                    
                    formula_dict = json.loads(molecularFormulaTable.mol_formula)
                    
                    if adduct_atom in formula_dict.keys():
                        formula_dict[adduct_atom] += 1  
                    else:
                        formula_dict[adduct_atom] = 1      
                    
                    mz = adduct_atom_mass + molecularFormulaTable.mz
                    nm = int(mz)
                    
                    new_formul_obj = molform_model( **{"mol_formula" : json.dumps(formula_dict),
                                            "mz" : mz,
                                            "ion_type" : ion_type,
                                            "nominal_mz" : nm,
                                            "ion_charge" : molecularFormulaTable.ion_charge,
                                            "classe" : molecularFormulaTable.classe,
                                            "C" : molecularFormulaTable.C,
                                            "H" : molecularFormulaTable.H,
                                            "N" : molecularFormulaTable.N,
                                            "O" : molecularFormulaTable.O,
                                            "S" : molecularFormulaTable.S,
                                            "P" : molecularFormulaTable.P,
                                            "H_C" : molecularFormulaTable.H_C,
                                            "O_C" : molecularFormulaTable.O_C,
                                            "DBE" : molecularFormulaTable.DBE,
                                            })
                    if nm in new_dict:
                        new_dict[nm].append(new_formul_obj)
                    
                    else:
                        new_dict[nm]= [new_formul_obj]
                    
        return new_dict




from matplotlib import pyplot as plt
import numpy as np

from corems.encapsulation.constant import Labels
from corems.encapsulation.factory.parameters import default_parameters
from corems.mass_spectrum.factory.MassSpectrumClasses import MassSpecCentroidLowRes
from corems.chroma_peak.factory.ChromaPeakClasses import GCPeakDeconvolved
from corems.mass_spectra.calc import SignalProcessing as sp

class MassDeconvolution:

    def run_deconvolution(self, plot_res=False):

        eic_dict = self.ion_extracted_chroma(self._ms)

        peaks_entity_data = self.find_peaks_entity(eic_dict)

        ''' select model peaks, create Mass Spectrum objs, GCPeak objs, store results in GC_Class gcpeaks obj'''
        self.deconvolution(peaks_entity_data, plot_res)

    def centroid_detector(self, tic, rt):
        ''' this function has been replaced with sp.peak_picking_first_derivative
            and it not used
        '''
        noise_std = self.chromatogram_settings.std_noise_threshold

        method = self.chromatogram_settings.noise_threshold_method

        ''' peak picking'''
        min_height = self.chromatogram_settings.peak_height_min_percent
        min_datapoints = self.chromatogram_settings.min_peak_datapoints

        ''' baseline detection'''
        max_prominence = self.chromatogram_settings.peak_max_prominence_percent
        max_height = self.chromatogram_settings.peak_height_max_percent

        peak_indexes_generator = sp.peak_detector_generator(tic, noise_std, method, rt, max_height, min_height, max_prominence, min_datapoints)

        return peak_indexes_generator

    def ion_extracted_chroma(self, mass_spectra_obj):

        eic_dict = {}

        for scan_number, ms_obj in mass_spectra_obj.items():

            mz_list = ms_obj.mz_exp
            abundance_list = ms_obj.abundance
            # add list of scan numbers
            for index, mz in enumerate(mz_list):

                # dict of mz and tuple (mass spectrum abundances index, and scan number)
                if mz not in eic_dict.keys():

                    eic_dict[mz] = [[abundance_list[index]], [ms_obj.rt] ]

                else:

                    eic_dict[mz][0].append(ms_obj.abundance[index])
                    eic_dict[mz][1].append(ms_obj.rt)

        return eic_dict

    def hc(self, X, Y, max_rt_distance=0.025):

        from scipy.cluster.hierarchy import dendrogram, linkage
        from scipy.cluster.hierarchy import fcluster
        # from matplotlib import pyplot as plt

        Z = linkage(np.reshape(X, (len(X), 1)), method="ward")
        # Z = linkage(X, method  = "ward")

        max_d = max_rt_distance
        distance_clusters = fcluster(Z, max_d, criterion='distance')
        # print("distance")
        # print(distance_clusters)

        # inconsistency_cluster = fcluster(Z, 2, depth=2)
        # max_cluster = fcluster(Z, 2, criterion='maxclust')

        grouped_rt = {}

        for index_obj, group in enumerate(distance_clusters):

            if group not in grouped_rt.keys():
                grouped_rt[group] = [X[index_obj]]
            else:
                grouped_rt[group].append(X[index_obj])

        # print(distance_clusters, grouped_rt)
        return grouped_rt      

        # plt.figure(figsize=(10, 8))
        # plt.scatter(X, Y, c=distance_clusters, cmap='prism')  # plot points with cluster dependent colors
        # plt.show()
        # labelList = range(int(min(X)), int(max(X)))

        # plt.figure(figsize=(10, 7))
        # dendrogram(Z,
        #            orientation='top',
        #            distance_sort='descending',
        #            show_leaf_counts=True)
        # plt.show()
        # print(Z)

    def find_peaks_entity(self, eic_dict):

        ''' combine eic with mathing rt apexes'''
        max_prominence = self.chromatogram_settings.peak_max_prominence_percent

        max_height = self.chromatogram_settings.peak_height_max_percent

        signal_threshold = self.chromatogram_settings.eic_signal_threshold

        min_peak_datapoints = self.chromatogram_settings.min_peak_datapoints

        correct_baseline = False
        peaks_entity_data = {}

        max_eic = 0
        for mz, eic_scan_index_rt in eic_dict.items():

            ind_max_eic = max(eic_scan_index_rt[0])
            max_eic = ind_max_eic if ind_max_eic > max_eic else max_eic

        for mz, eic_scan_index_rt in eic_dict.items():

            eic = eic_scan_index_rt[0]
            rt_list = eic_scan_index_rt[1]

            if len(eic) >= min_peak_datapoints:

                smooth_eic = self.smooth_tic(eic)

                include_indexes = sp.peak_picking_first_derivative(rt_list, smooth_eic, max_height, max_prominence, max_eic, min_peak_datapoints,
                                                                   signal_threshold=signal_threshold, correct_baseline=correct_baseline)

                for initial_scan, apex_scan, final_scan in include_indexes:

                    rt_corrected_therm = self.quadratic_interpolation(rt_list, smooth_eic, apex_scan)

                    ref_apex_rt = round(rt_list[apex_scan] + rt_corrected_therm, 4)

                    apex_rt = rt_list[apex_scan]
                    # apex_abundance = smooth_eic[apex_scan]

                    # maximum_tic = apex_abundance if apex_abundance > maximum_tic else maximum_tic

                    for scan_index in range(initial_scan, final_scan):

                        peak_rt = rt_list[scan_index]
                        peak_abundance = smooth_eic[scan_index]

                        if peak_abundance > 0:

                            dict_data = {peak_rt: {'mz': [mz],
                                                    'abundance': [peak_abundance],
                                                    'scan_number': [scan_index]},
                                                    'ref_apex_rt': ref_apex_rt
                                                    }

                            if apex_rt not in peaks_entity_data.keys():

                                peaks_entity_data[apex_rt] = dict_data

                            else:

                                if peak_rt not in peaks_entity_data[apex_rt].keys():

                                    peaks_entity_data[apex_rt][peak_rt] = dict_data.get(peak_rt)

                                else:

                                    existing_data = peaks_entity_data[apex_rt].get(peak_rt)

                                    existing_data['mz'].append(mz)
                                    existing_data['abundance'].append(peak_abundance)
                                    existing_data['scan_number'].append(scan_index)


        return peaks_entity_data

    def mass_spec_factory(self, rt, datadict):

        # tic = sum(datadict.get('abundance'))

        scan_index = datadict['scan_number'][0]

        mz_list, abundance_list = zip(*sorted(zip(datadict['mz'], datadict['abundance'])))

        data_dict = {Labels.mz: mz_list, Labels.abundance: abundance_list}

        d_params = default_parameters(self._ms[scan_index]._filename)

        d_params["rt"] = rt

        d_params["scan_number"] = scan_index

        d_params['label'] = Labels.gcms_centroid

        d_params["polarity"] = self._ms[scan_index].polarity

        d_params['analyzer'] = self._ms[scan_index].analyzer

        d_params['instrument_label'] = self._ms[scan_index].instrument_label

        d_params["filename_path"] = self._ms[scan_index].instrument_label

        ms = MassSpecCentroidLowRes(data_dict, d_params)

        return ms

    def smooth_signal(self, signal):

        implemented_smooth_method = self.chromatogram_settings.implemented_smooth_method

        pol_order = self.chromatogram_settings.savgol_pol_order

        window_len = self.chromatogram_settings.smooth_window

        window = self.chromatogram_settings.smooth_method

        return sp.smooth_signal(signal, window_len, window, pol_order, implemented_smooth_method)

    def add_gcpeak(self, new_apex_index, start_rt, final_rt, peak_rt, smoothed_tic, datadict, plot_res):

        if start_rt <= peak_rt[new_apex_index[1]] <= final_rt:

            rt_list = peak_rt[new_apex_index[0]:new_apex_index[2]]
            tic_list = smoothed_tic[new_apex_index[0]:new_apex_index[2]]

            apex_rt = peak_rt[new_apex_index[1]]
            apex_i = rt_list.index(apex_rt)

            '''workaround for peak picking missing some local minimas'''
            if not apex_rt in self.processed_appexes:

                self.processed_appexes.append(apex_rt)

                mass_spectra = (self.mass_spec_factory(rt, datadict.get(rt)) for rt in rt_list)

                gc_peak = GCPeakDeconvolved(self, mass_spectra, apex_i, rt_list, tic_list)

                gc_peak.calc_area(tic_list, 1)

                self.gcpeaks.append(gc_peak)

                if plot_res:

                    plt.plot(gc_peak.rt_list, gc_peak.tic_list)
                    plt.plot(gc_peak.rt, gc_peak.tic, c='black', marker= '^', linewidth=0)

    def deconvolution(self, peaks_entity_data, plot_res):

        # plot_res = True 
        domain = self.retention_time
        signal = self._processed_tic
        max_height = self.chromatogram_settings.peak_height_max_percent
        max_prominence = self.chromatogram_settings.peak_max_prominence_percent
        min_peak_datapoints = self.chromatogram_settings.min_peak_datapoints
        signal_threshold = self.chromatogram_settings.peak_height_min_percent
        max_rt_distance = self.chromatogram_settings.max_rt_distance

        max_signal = max(signal)
        correct_baseline = False

        include_indexes = sp.peak_picking_first_derivative(domain, signal, max_height, max_prominence, max_signal, min_peak_datapoints,
                                                           signal_threshold=signal_threshold, correct_baseline=correct_baseline, plot_res=False)

        ''' deconvolution window is defined by the TIC peak region'''
        all_apexes_rt = np.array(list(peaks_entity_data.keys()))

        '''workaround for peak picking missing some local minimas'''
        self.processed_appexes = []

        for indexes_tuple in include_indexes:

            start_rt = self.retention_time[indexes_tuple[0]]
            # apex_rt = self.retention_time[indexes_tuple[1]]
            final_rt = self.retention_time[indexes_tuple[2]]

            ''' find all features within TIC peak window'''
            peak_features_indexes = np.where((all_apexes_rt > start_rt) & (all_apexes_rt < final_rt))[0]
            peak_features_rts = all_apexes_rt[peak_features_indexes]

            # print(start_rt, apex_rt, final_rt )

            filtered_features_rt = []
            filtered_features_abundance = []

            for each_apex_rt in peak_features_rts:

                apex_data = peaks_entity_data.get(each_apex_rt).get(each_apex_rt)

                peak_features_tic = sum(peaks_entity_data.get(each_apex_rt).get(each_apex_rt).get('abundance'))

                norm_smooth_tic = (peak_features_tic / max_signal) * 100

                ''' TODO: 
                    Improve Peak Filtering

                    Calculate peaks sharpness here and filter it out (Amax - An /n)?
                    Peak Fit and Calculate Peak Gaussian Similarity?
                    Currentely using flat % tic relative abundance threshold and min 3 m/z per mass spectrum
                '''
                if norm_smooth_tic > signal_threshold and len(apex_data['mz']) > 1:

                    # print(len(apex_data['mz']))
                    filtered_features_rt.append(each_apex_rt)
                    filtered_features_abundance.append(peak_features_tic)

            if len(filtered_features_rt) > 1:
                ''' more than one peak feature identified inside a TIC peak  '''
                # plt.plot(self.retention_time[indexes_tuple[0]:indexes_tuple[2]], signal[indexes_tuple[0]:indexes_tuple[2]], c='black')

                # print(filtered_features_rt)
                grouped_rt = self.hc(filtered_features_rt, filtered_features_abundance, max_rt_distance=max_rt_distance)
                # print(grouped_rt)

                for group, apex_rt_list in grouped_rt.items():
                    ''' each group is a peak feature defined by the hierarchical clutter algorithm

                    '''
                    group_datadict = {}
                    group_datadict['ref_apex_rt'] = []

                    for each_group_apex_rt in apex_rt_list:

                        datadict = peaks_entity_data.get(each_group_apex_rt)

                        for rt, each_datadict in datadict.items():

                            if rt == "ref_apex_rt":

                                group_datadict['ref_apex_rt'].append(each_datadict)

                            else:

                                if rt in group_datadict.keys():

                                    mz_list = each_datadict.get("mz")
                                    abundance_list = each_datadict.get("abundance")

                                    each_mz_abun = dict(zip(mz_list, abundance_list)) 

                                    for index_mz, mz in enumerate(group_datadict[rt].get("mz")):
                                        if mz in each_mz_abun.keys():

                                            each_mz_abun[mz] = each_mz_abun[mz] + group_datadict[rt].get("abundance")[index_mz]

                                        else:

                                            each_mz_abun[mz] = group_datadict[rt].get("abundance")[index_mz]

                                    group_datadict[rt] = {'mz': list(each_mz_abun.keys()),
                                                          'abundance': list(each_mz_abun.values()),
                                                          'scan_number': each_datadict.get('scan_number')}

                                else:

                                    group_datadict[rt] = each_datadict

                    peak_rt = []
                    peak_tic = []

                    # print(group_datadict.get('ref_apex_rt'))
                    for rt, each_datadict in group_datadict.items():
                        if rt != "ref_apex_rt":
                            peak_rt.append(rt)
                            peak_tic.append(sum(each_datadict["abundance"]))

                    peak_rt, peak_tic = zip(*sorted(zip(peak_rt, peak_tic)))

                    smoothed_tic = self.smooth_signal(peak_tic)

                    include_indexes = sp.peak_picking_first_derivative(peak_rt, smoothed_tic,  max_height, max_prominence, max_signal, min_peak_datapoints,
                                                                            signal_threshold=signal_threshold,  correct_baseline=False, plot_res=False)

                    include_indexes = list(include_indexes)

                    if include_indexes:

                        if len(include_indexes) > 1:
                            ''' after sum there are two apexes
                                check if it is inside the deconvolution window, otherwise ignores it
                            '''

                            for new_apex_index in include_indexes:
                                # pass
                                self.add_gcpeak(new_apex_index, start_rt, final_rt, peak_rt, smoothed_tic, group_datadict, plot_res)

                        else:
                            ''' after sum there is on apex
                                save it
                            ''' 
                            new_apex_index = include_indexes[0]
                            # print(include_indexes, group, apex_rt_list)
                            self.add_gcpeak(new_apex_index, start_rt, final_rt, peak_rt, smoothed_tic, group_datadict, plot_res)

            elif len(filtered_features_rt) == 1:
                ''' only one peak feature inside deconvolution window '''

                each_apex_rt = filtered_features_rt[0]

                datadict = peaks_entity_data.get(each_apex_rt)

                peak_rt = []
                peak_tic = []

                for rt, each_datadict in datadict.items():

                    if rt != "ref_apex_rt":
                        peak_rt.append(rt)
                        peak_tic.append(sum(each_datadict["abundance"]))

                peak_rt, peak_tic = zip(*sorted(zip(peak_rt, peak_tic)))

                smoothed_tic = self.smooth_signal(peak_tic)

                include_indexes = sp.peak_picking_first_derivative(peak_rt, smoothed_tic,  max_height, max_prominence, max_signal, min_peak_datapoints,
                                                                            signal_threshold=signal_threshold,  correct_baseline=False, plot_res=False)
                include_indexes = list(include_indexes)   

                if include_indexes:

                        ''' after sum there are two apexes
                            check if it is inside the deconvolution window, otherwise ignores it'''
                        if len(include_indexes) > 1:

                            for new_apex_index in include_indexes:
                                # pass
                                self.add_gcpeak(new_apex_index, start_rt, final_rt, peak_rt, smoothed_tic, datadict, plot_res)

                        else:
                            ''' after sum there is one apex
                            save it
                            includes_indexes = (start, apex, final )'''

                            new_apex_index = include_indexes[0]

                            self.add_gcpeak(new_apex_index, start_rt, final_rt, peak_rt, smoothed_tic, datadict, plot_res)

            else:

                # print('no data after filter')
                pass
        if plot_res:
            plt.plot(self.retention_time, self._processed_tic, c='black')
            plt.show()

    def quadratic_interpolation(self, rt_list, tic_list, apex_index):

        rt_list = np.array(rt_list)
        tic_list = np.array(tic_list)
        three_highest_i = [i for i in range(apex_index - 1, apex_index + 2)]

        z = np.poly1d(np.polyfit(rt_list[three_highest_i], tic_list[three_highest_i], 2))
        a = z[2]
        b = z[1]

        corrected_apex_rt = -b / (2 * a)
        initial_rt = rt_list[apex_index]

        return initial_rt - corrected_apex_rt
